"""Test generation of net charge calculation."""

import unittest

import numpy as np
import pandas as pd

from molpipeline import ErrorFilter, FilterReinserter, Pipeline
from molpipeline.any2mol import SmilesToMol
from molpipeline.mol2any import MolToNetCharge

DF_TEST_DATA = pd.DataFrame(
    {
        "smiles": [
            "[Fe+2]",
            "c1cc(c(nc1)Cl)C(=O)Nc2c(c3c(s2)CCCCC3)C(=O)N",
            "Cc1ccc(cc1)S(=O)(=O)Nc2c(c3c(s2)C[C@@H](CC3)C)C(=O)N",
            "c1cc(oc1)CN=C2C=C(C(=CC2=C(O)[O-])S(=O)(=O)[NH-])Cl",
            "C[C@@H]1[C@@H](OP2(O1)(O[C@H]([C@H](O2)C)C)C[NH+]3CCCCC3)C",  # this one fails gasteiger charge computation
        ],
        "expected_net_charges_formal_charge": [2, 0, 0, -2, 1],
        "expected_net_charges_gasteiger": [2, -1, -1, -2, np.nan],
    }
)


class TestNetChargeCalculator(unittest.TestCase):
    """Unittest for MolToNetCharge, which calculates net charges of molecules."""

    def test_net_charge_calculation_formal_charge(self) -> None:
        """Test if the net charge calculation works as expected for formal charges.

        Returns
        -------
        None
        """
        # we need the error filter and reinserter to handle the case where the charge calculation fails
        error_filter = ErrorFilter(filter_everything=True)
        pipeline = Pipeline(
            [
                ("smi2mol", SmilesToMol()),
                (
                    "net_charge_element",
                    MolToNetCharge(charge_method="formal_charge", standardizer=None),
                ),
                ("error_filter", error_filter),
                (
                    "filter_reinserter",
                    FilterReinserter.from_error_filter(error_filter, fill_value=np.nan),
                ),
            ],
        )

        actual_net_charges = pipeline.fit_transform(DF_TEST_DATA["smiles"])
        self.assertTrue(
            np.allclose(
                DF_TEST_DATA["expected_net_charges_formal_charge"]
                .to_numpy()
                .reshape(-1, 1),
                actual_net_charges,
                equal_nan=True,
            )
        )

    def test_net_charge_calculation_gasteiger(self) -> None:
        """Test if the net charge calculation works as expected for gasteiger charges.

        Returns
        -------
        None
        """
        # we need the error filter and reinserter to handle the case where the charge calculation fails
        error_filter = ErrorFilter(filter_everything=True)
        pipeline = Pipeline(
            [
                ("smi2mol", SmilesToMol()),
                (
                    "net_charge_element",
                    MolToNetCharge(charge_method="gasteiger", standardizer=None),
                ),
                ("error_filter", error_filter),
                (
                    "filter_reinserter",
                    FilterReinserter.from_error_filter(error_filter, fill_value=np.nan),
                ),
            ],
        )

        actual_net_charges = pipeline.fit_transform(DF_TEST_DATA["smiles"])
        self.assertTrue(
            np.allclose(
                DF_TEST_DATA["expected_net_charges_gasteiger"]
                .to_numpy()
                .reshape(-1, 1),
                actual_net_charges,
                equal_nan=True,
            )
        )


if __name__ == "__main__":
    unittest.main()



"""Connected component clustering algorithm."""

from __future__ import annotations

import numpy as np
import numpy.typing as npt
from scipy.sparse import csr_matrix

from molpipeline.estimators.algorithm.union_find import UnionFindNode
from molpipeline.utils.kernel import tanimoto_similarity_sparse


def calc_chunk_size_from_memory_requirement(
    nof_rows: int, nof_cols: int, itemsize: int, memory_cutoff: float
) -> int:
    """Calculate the chunk size from the memory requirement.

    Parameters
    ----------
    nof_rows: int
        Number of rows of the matrix.
    nof_cols: int
        Number of columns of the matrix.
    itemsize: int
        Itemsize of the matrix.
    memory_cutoff: float
        Memory cutoff in GB.

    Returns
    -------
    int
        Chunk size in number of rows.
    """
    memory_cutoff_byte: float = memory_cutoff * 1024**3
    # get memory requirement of dense matrix
    row_memory: int = nof_cols * itemsize
    # get allowed rows
    allowed_rows = max(int(memory_cutoff_byte / row_memory), 1)
    allowed_rows = min(allowed_rows, nof_rows)
    return allowed_rows


def connected_components_iterative_algorithm(
    feature_mat: csr_matrix, similarity_threshold: float, chunk_size: int = 5000
) -> tuple[int, npt.NDArray[np.int32]]:
    """Compute connected component clustering iteratively.

    This algorithm is suited for large data sets since the complete similarity matrix is not stored in memory at once.

    Parameters
    ----------
    feature_mat: csr_matrix
        Feature matrix from which to calculate row-wise similarities.
    similarity_threshold: float
        Similarity threshold used to determine edges in the graph representation.
    chunk_size: int
        Number of rows for which similarities are determined at one iteration of the algorithm.

    Returns
    -------
    tuple[int, np.ndarray[int]]
        Number of clusters and cluster labels.
    """
    nof_samples = feature_mat.shape[0]
    uf_nodes = [UnionFindNode() for _ in range(nof_samples)]

    for i in range(0, nof_samples, chunk_size):
        mat_chunk = feature_mat[i : i + chunk_size, :]

        similarity_mat_chunk = tanimoto_similarity_sparse(mat_chunk, feature_mat)

        indices = np.transpose(
            np.asarray(similarity_mat_chunk >= similarity_threshold).nonzero()
        )

        for i_idx, j_idx in indices:
            if i + i_idx >= j_idx:
                continue
            uf_nodes[j_idx].union(uf_nodes[i + i_idx])

    return UnionFindNode.get_connected_components(uf_nodes)



"""Nearest neighbor algorithm."""

from __future__ import annotations

from typing import Any, Callable, Literal, Sequence, Union

try:
    from typing import Self
except ImportError:
    from typing_extensions import Self


import numpy as np
import numpy.typing as npt
from scipy import sparse
from sklearn.neighbors import NearestNeighbors

from molpipeline.utils.value_checks import get_length

__all__ = ["NamedNearestNeighbors"]

Algorithm = Literal["auto", "ball_tree", "kd_tree", "brute"]

SklearnNativeMetrics = Literal[
    "cityblock",
    "cosine",
    "euclidean",
    "haversine",
    "jaccard",
    "l1",
    "l2",
    "manhattan",
    "minkowski",
    "nan_euclidean",
    "precomputed",
]

AllMetrics = Union[
    SklearnNativeMetrics,
    Callable[[Any, Any], float | npt.NDArray[np.float64] | Sequence[float]],
]


class NamedNearestNeighbors(NearestNeighbors):  # pylint: disable=too-many-ancestors
    """NearestNeighbors with a name attribute."""

    learned_names_: npt.NDArray[Any] | None

    def __init__(
        self,
        n_neighbors: int = 5,
        radius: float = 1.0,
        algorithm: Algorithm = "auto",
        leaf_size: int = 30,
        metric: AllMetrics = "minkowski",
        p: int = 2,
        metric_params: dict[str, Any] | None = None,
        n_jobs: int | None = None,
    ):
        """Initialize the nearest neighbor algorithm.

        Parameters
        ----------
        n_neighbors : int, optional (default = 5)
            The number of neighbors to get.
        radius : float, optional (default = 1.0)
            Range of parameter space to use by default for radius_neighbors queries.
        algorithm : {'auto', 'ball_tree', 'kd_tree', 'brute'}, optional (default = 'auto')
            Algorithm used to compute the nearest neighbors.
        leaf_size : int, optional (default = 30)
            Leaf size passed to BallTree or KDTree. This can affect the speed of the construction and query,
            as well as the memory required to store the tree. The optimal value depends on the nature of the problem.
        metric : Union[str, Callable], optional (default = 'minkowski')
            The distance metric to use for the tree.
            The default metric is minkowski, and with p=2 is equivalent to the standard Euclidean metric.
        p : int, optional (default = 2)
            Power parameter for the Minkowski metric.
        metric_params : dict, optional (default = None)
            Additional keyword arguments for the metric function.
        n_jobs : int, optional (default = None)
            The number of parallel jobs to run for neighbors search. None means 1 unless in a joblib.parallel_backend context.
            -1 means using all processors.
        """
        super().__init__(
            n_neighbors=n_neighbors,
            radius=radius,
            algorithm=algorithm,
            leaf_size=leaf_size,
            metric=metric,
            p=p,
            metric_params=metric_params,
            n_jobs=n_jobs,
        )
        self.learned_names_ = None

    # pylint: disable=arguments-differ, signature-differs
    def fit(
        self,
        X: (
            npt.NDArray[Any] | sparse.csr_matrix | Sequence[Any]
        ),  # pylint: disable=invalid-name
        y: Sequence[Any],  # pylint: disable=invalid-name
    ) -> Self:
        """Fit the model using X as training data.

        Parameters
        ----------
        X : npt.NDArray[Any] | sparse.csr_matrix | Sequence[Any]
            Training data.
        y : Sequence[Any]
            Target values. Here values are used as returned nearest neighbors.
            Must have the same length as X.
            Will be stored as the learned_names_ attribute as npt.NDArray[Any].

        Returns
        -------
        Self
            The instance itself.

        Raises
        ------
        ValueError
            If the input arrays have different lengths or do not have a shape nor len attribute.
        """
        # Check if X and y have the same length
        n_x = get_length(X)  # Allowing for any sequence type
        n_y = get_length(y)
        if n_x != n_y:
            raise ValueError("X and y must have the same length.")

        self.learned_names_ = np.array(y)
        super().fit(X)
        return self

    # pylint: disable=invalid-name
    def predict(
        self,
        X: npt.NDArray[Any] | sparse.csr_matrix | Sequence[Any],
        return_distance: bool = False,
        n_neighbors: int | None = None,
    ) -> npt.NDArray[Any]:
        """Find the k-neighbors of a point.

        Parameters
        ----------
        X : npt.NDArray[Any] | sparse.csr_matrix
            The new data to query.
        return_distance : bool, optional (default = False)
            If True, return the distances to the neighbors of each point.
            Default: False
        n_neighbors : int, optional (default = None)
            Number of neighbors to get. If None, the value set at initialization is used.

        Returns
        -------
        tuple[npt.NDArray[Any], npt.NDArray[np.float64]] | npt.NDArray[Any]
            The indices of the nearest points in the population matrix and the distances to the points.
        """
        if self.learned_names_ is None:
            raise ValueError("The model has not been fitted yet.")
        if n_neighbors is None:
            n_neighbors = self.n_neighbors

        if return_distance:
            distances, indices = super().kneighbors(
                X, n_neighbors=n_neighbors, return_distance=True
            )
            # stack in such a way that the shape is (n_input, n_neighbors, 2)
            # shape 2 as the neighbor idx and distance are returned
            r_arr = np.stack([self.learned_names_[indices], distances], axis=2)
            return r_arr

        indices = super().kneighbors(X, n_neighbors=n_neighbors, return_distance=False)
        return self.learned_names_[indices]

    def fit_predict(
        self,
        X: npt.NDArray[Any] | sparse.csr_matrix,  # pylint: disable=invalid-name
        y: Sequence[Any],
        return_distance: bool = False,
        n_neighbors: int | None = None,
    ) -> tuple[npt.NDArray[Any], npt.NDArray[np.float64]] | npt.NDArray[Any]:
        """Find the k-neighbors of a point.

        Parameters
        ----------
        X : npt.NDArray[Any] | sparse.csr_matrix
            The new data to query.
        y : Sequence[Any]
            Target values. Here values are used as returned nearest neighbors.
            Must have the same length as X.
        return_distance : bool, optional (default = False)
            If True, return the distances to the neighbors of each point.
            Default: False
        n_neighbors : int, optional (default = None)
            Number of neighbors to get. If None, the value set at initialization is used.

        Returns
        -------
        Tuple[array, array]
            The indices of the nearest points in the population matrix and the distances to the points.
        """
        self.fit(X, y)
        return self.predict(X, return_distance=return_distance, n_neighbors=n_neighbors)



"""Classes ment to transform given inchi to a RDKit molecule."""

from rdkit import Chem

from molpipeline.abstract_pipeline_elements.any2mol.string2mol import (
    SimpleStringToMolElement,
)
from molpipeline.utils.molpipeline_types import RDKitMol


class InchiToMol(SimpleStringToMolElement):
    """Transforms Inchi to RDKit Mol objects."""

    def string_to_mol(self, value: str) -> RDKitMol:
        """Transform Inchi string to molecule.

        Parameters
        ----------
        value: str
            Inchi string.

        Returns
        -------
        RDKitMol
            Rdkit molecule if valid Inchi, else None.
        """
        return Chem.MolFromInchi(value)



"""Test leader picker clustering estimator."""

from __future__ import annotations

import unittest
from typing import Any

import numpy as np
from rdkit import DataStructs

from molpipeline import Pipeline
from molpipeline.any2mol import AutoToMol
from molpipeline.estimators import LeaderPickerClustering
from molpipeline.mol2any import MolToMorganFP


class TestLeaderPickerEstimator(unittest.TestCase):
    """Test LeaderPicker clustering estimator."""

    def test_leader_picker_clustering_estimator(self) -> None:
        """Test LeaderPicker clustering estimator."""

        fingerprint_matrix = [
            DataStructs.CreateFromBitString(x)
            for x in [
                "000",  # 0
                "100",  # 1
                "110",  # 2
                "101",  # 3
                "010",  # 4
                "011",  # 5
                "001",  # 6
                "111",  # 7
                "000",
                "100",
                "110",
                "101",
                "010",
                "011",
                "001",
                "111",
            ]
        ]

        eps: float = 1e-10

        expected_clusterings: list[dict[str, Any]] = [
            {
                "threshold": 0.0,
                "expected_clustering": [0, 1, 2, 3, 4, 5, 6, 7, 0, 1, 2, 3, 4, 5, 6, 7],
            },
            {
                "threshold": 0.0 + eps,
                "expected_clustering": [0, 1, 2, 3, 4, 5, 6, 7, 0, 1, 2, 3, 4, 5, 6, 7],
            },
            {
                "threshold": 1 / 3 - eps,
                "expected_clustering": [0, 1, 2, 3, 4, 5, 6, 7, 0, 1, 2, 3, 4, 5, 6, 7],
            },
            {
                "threshold": 1 / 3,
                "expected_clustering": [0, 1, 2, 3, 4, 5, 6, 7, 0, 1, 2, 3, 4, 5, 6, 7],
            },
        ]

        for test_case_dict in expected_clusterings:
            threshold = test_case_dict["threshold"]
            expected_clustering = test_case_dict["expected_clustering"]

            self.assertEqual(len(fingerprint_matrix), len(expected_clustering))

            estimator = LeaderPickerClustering(distance_threshold=threshold)
            actual_labels = estimator.fit_predict(fingerprint_matrix)

            exp_nof_clusters = np.unique(expected_clustering).shape[0]
            self.assertEqual(exp_nof_clusters, estimator.n_clusters_)

            self.assertTrue(np.equal(actual_labels, expected_clustering).all())

    def test_leader_picker_pipeline(self) -> None:
        """Test leader picker clustering in pipeline."""

        test_smiles = ["C", "N", "c1ccccc1", "c1ccc(O)cc1", "CCCCCCO", "CCCCCCC"]

        distances = [0.05, 0.95]
        expected_labels = [[0, 1, 2, 3, 4, 5], [0, 1, 2, 2, 3, 3]]
        expected_centroids = [[0, 1, 2, 3, 4, 5], [0, 1, 2, 4]]

        for dist, exp_labels, exp_centroids in zip(
            distances, expected_labels, expected_centroids
        ):

            leader_picker = LeaderPickerClustering(distance_threshold=dist)
            pipeline = Pipeline(
                [
                    ("auto2mol", AutoToMol()),
                    (
                        "morgan2",
                        MolToMorganFP(
                            return_as="explicit_bit_vect", n_bits=1024, radius=2
                        ),
                    ),
                    ("leader_picker", leader_picker),
                ],
            )

            actual_labels = pipeline.fit_predict(test_smiles)

            self.assertTrue(np.equal(actual_labels, exp_labels).all())
            self.assertIsNotNone(leader_picker.centroids_)
            self.assertTrue(np.equal(leader_picker.centroids_, exp_centroids).all())  # type: ignore



"""Explainer classes for explaining predictions."""

from __future__ import annotations

import abc
from typing import Any, Callable

import numpy as np
import numpy.typing as npt
import pandas as pd
import shap
from scipy.sparse import issparse, spmatrix
from sklearn.base import BaseEstimator

try:
    from typing import override  # type: ignore[attr-defined]
except ImportError:
    from typing_extensions import override

from molpipeline import Pipeline
from molpipeline.abstract_pipeline_elements.core import OptionalMol
from molpipeline.experimental.explainability.explanation import (
    AtomExplanationMixin,
    BondExplanationMixin,
    FeatureExplanationMixin,
    FeatureInfoMixin,
    SHAPExplanationMixin,
    SHAPFeatureAndAtomExplanation,
    SHAPFeatureExplanation,
)
from molpipeline.experimental.explainability.fingerprint_utils import (
    fingerprint_shap_to_atomweights,
)
from molpipeline.mol2any import MolToMorganFP
from molpipeline.utils.subpipeline import SubpipelineExtractor, get_model_from_pipeline


def _to_dense(
    feature_matrix: npt.NDArray[Any] | spmatrix,
) -> npt.NDArray[Any]:
    """Mitigate feature incompatibility with SHAP objects.

    Parameters
    ----------
    feature_matrix : npt.NDArray[Any] | spmatrix
        The input features.

    Returns
    -------
    Any
        The input features in a compatible format.
    """
    if issparse(feature_matrix):
        return feature_matrix.todense()  # type: ignore[union-attr]
    return feature_matrix


def _get_prediction_function(
    pipeline: Pipeline | BaseEstimator,
) -> Callable[[npt.ArrayLike], npt.ArrayLike]:
    """Get the prediction function of a model.

    Parameters
    ----------
    pipeline : Pipeline
        The pipeline containing the model.

    Returns
    -------
    Any
        The prediction function.
    """
    if hasattr(pipeline, "predict_proba"):
        return pipeline.predict_proba
    if hasattr(pipeline, "decision_function"):
        return pipeline.decision_function
    if hasattr(pipeline, "predict"):
        return pipeline.predict
    raise ValueError("Could not determine the model output predictions")


# This function might also be put at a more central position in the lib.
def _get_predictions(
    pipeline: Pipeline, feature_matrix: npt.NDArray[Any] | spmatrix
) -> npt.NDArray[np.float64]:
    """Get the predictions of a model.

    Raises if no adequate method is found.

    Parameters
    ----------
    pipeline : Pipeline
        The pipeline containing the model.
    feature_matrix : Any
        The input data.

    Returns
    -------
    npt.NDArray[np.float64]
        The predictions.
    """
    prediction_function = _get_prediction_function(pipeline)
    prediction = prediction_function(feature_matrix)
    return np.array(prediction)


def _convert_shap_feature_weights_to_atom_weights(
    feature_weights: npt.NDArray[np.float64],
    molecule: OptionalMol,
    featurization_element: MolToMorganFP,
    feature_vector: npt.NDArray[np.float64],
) -> npt.NDArray[np.float64]:
    """Convert SHAP feature weights to atom weights.

    Parameters
    ----------
    feature_weights : npt.NDArray[np.float64]
        The feature weights.
    molecule : OptionalMol
        The molecule.
    featurization_element : MolToMorganFP
        The featurization element.
    feature_vector : npt.NDArray[np.float64]
        The feature vector.

    Returns
    -------
    npt.NDArray[np.float64]
        The atom weights.
    """
    if feature_weights.ndim == 1:
        # regression case
        feature_weights_present_bits_only = feature_weights.copy()
    elif feature_weights.ndim == 2:
        # binary classification case. Take the weights for the positive class.
        feature_weights_present_bits_only = feature_weights[:, 1].copy()
    else:
        raise ValueError(
            "Unsupported number of dimensions for feature weights. Expected 1 or 2."
        )

    # reset shap values for bits that are not present in the molecule
    feature_weights_present_bits_only[feature_vector == 0] = 0

    atom_weights = np.array(
        fingerprint_shap_to_atomweights(
            molecule,
            featurization_element,
            feature_weights_present_bits_only,
        )
    )
    return atom_weights


class AbstractSHAPExplainer(abc.ABC):  # pylint: disable=too-few-public-methods
    """Abstract class for SHAP explainer objects."""

    @abc.abstractmethod
    def explain(
        self, X: Any, **kwargs: Any  # pylint: disable=invalid-name,unused-argument
    ) -> list[SHAPFeatureExplanation | SHAPFeatureAndAtomExplanation]:
        """Explain the predictions for the input data.

        Parameters
        ----------
        X : Any
            The input data to explain.
        kwargs : Any
            Additional keyword arguments.

        Returns
        -------
        list[SHAPFeatureExplanation | SHAPFeatureAndAtomExplanation]
            List of explanations corresponding to the input samples.
        """


class SHAPExplainerAdapter(
    AbstractSHAPExplainer, abc.ABC
):  # pylint: disable=too-few-public-methods
    """Adapter for SHAP explainer wrappers for handling molecules and pipelines."""

    # used for dynamically defining the return type of the explain method
    return_element_type_: type[SHAPFeatureExplanation | SHAPFeatureAndAtomExplanation]

    def __init__(
        self,
        pipeline: Pipeline,
        explainer: shap.TreeExplainer | shap.KernelExplainer,
    ) -> None:
        """Initialize the SHAPTreeExplainer.

        Parameters
        ----------
        pipeline : Pipeline
            The pipeline containing the model to explain.
        explainer : shap.TreeExplainer | shap.KernelExplainer
            The shap explainer object.
        """
        self.pipeline = pipeline
        self.explainer = explainer

        pipeline_extractor = SubpipelineExtractor(self.pipeline)

        # extract the molecule reader subpipeline
        self.molecule_reader_subpipeline = (
            pipeline_extractor.get_molecule_reader_subpipeline()
        )
        if self.molecule_reader_subpipeline is None:
            raise ValueError("Could not determine the molecule reader subpipeline.")

        # extract the featurization subpipeline
        self.featurization_subpipeline = (
            pipeline_extractor.get_featurization_subpipeline()
        )
        if self.featurization_subpipeline is None:
            raise ValueError("Could not determine the featurization subpipeline.")

        # determine type of returned explanation
        featurization_element = self.featurization_subpipeline.steps[-1][1]  # type: ignore[union-attr]
        if isinstance(featurization_element, MolToMorganFP):
            self.return_element_type_ = SHAPFeatureAndAtomExplanation
        else:
            self.return_element_type_ = SHAPFeatureExplanation

    @staticmethod
    def _prediction_is_valid(prediction: Any) -> bool:
        """Check if the prediction is valid using some heuristics.

        Can be used to catch inputs that failed the pipeline for some reason.

        Parameters
        ----------
        prediction : Any
            The prediction.
        Returns
        -------
        bool
            Whether the prediction is valid.
        """
        # if no prediction could be obtained (length is 0); the prediction guaranteed failed.
        if len(prediction) == 0:
            return False

        # use pandas.isna function to check for invalid predictions, e.g. None, np.nan,
        # pd.NA. Note that fill values like 0 will be considered as valid predictions.
        if pd.isna(prediction).any():
            return False

        return True

    @override
    def explain(
        self, X: Any, **kwargs: Any  # pylint: disable=invalid-name,unused-argument
    ) -> list[SHAPFeatureExplanation | SHAPFeatureAndAtomExplanation]:
        """Explain the predictions for the input data.

        If the calculation of the SHAP values for an input sample fails, the explanation will be invalid.
        This can be checked with the Explanation.is_valid() method.

        Parameters
        ----------
        X : Any
            The input data to explain.
        kwargs : Any
            Additional keyword arguments for SHAP's TreeExplainer.shap_values.

        Returns
        -------
        list[SHAPFeatureExplanation | SHAPFeatureAndAtomExplanation]
            List of explanations corresponding to the input data.
        """
        featurization_element = self.featurization_subpipeline.steps[-1][1]  # type: ignore[union-attr]

        explanation_results: list[
            SHAPFeatureExplanation | SHAPFeatureAndAtomExplanation
        ] = []
        for input_sample in X:

            input_sample = [input_sample]

            # get predictions
            prediction = _get_predictions(self.pipeline, input_sample)
            if not self._prediction_is_valid(prediction):
                # we use the prediction to check if the input is valid. If not, we cannot explain it.
                explanation_results.append(self.return_element_type_())
                continue

            if prediction.ndim > 1:
                prediction = prediction.squeeze()

            # get the molecule
            molecule = self.molecule_reader_subpipeline.transform(input_sample)[0]  # type: ignore[union-attr]

            # get feature vectors
            feature_vector = self.featurization_subpipeline.transform(input_sample)  # type: ignore[union-attr]
            feature_vector = _to_dense(feature_vector)
            feature_vector = np.asarray(feature_vector).squeeze()

            if feature_vector.size == 0:
                # if the feature vector is empty, we cannot explain the prediction.
                # This happens for failed instances in pipeline with fill values
                # that could be valid predictions, like 0.
                explanation_results.append(self.return_element_type_())
                continue

            # compute the shap values for the features
            feature_weights = self.explainer.shap_values(feature_vector, **kwargs)
            feature_weights = np.asarray(feature_weights).squeeze()

            atom_weights = None
            bond_weights = None

            if issubclass(
                self.return_element_type_, AtomExplanationMixin
            ) and isinstance(featurization_element, MolToMorganFP):
                # for Morgan fingerprint, we can map the shap values to atom weights
                atom_weights = _convert_shap_feature_weights_to_atom_weights(
                    feature_weights,
                    molecule,
                    featurization_element,
                    feature_vector,
                )

            # gather all input data for the explanation type to be returned
            explanation_data = {
                "molecule": molecule,
                "prediction": prediction,
            }
            if issubclass(self.return_element_type_, FeatureInfoMixin):
                explanation_data["feature_vector"] = feature_vector
                if not hasattr(featurization_element, "feature_names"):
                    raise ValueError(
                        "Featurization element does not have a get_feature_names method."
                    )
                explanation_data["feature_names"] = featurization_element.feature_names  # type: ignore[union-attr]

            if issubclass(self.return_element_type_, FeatureExplanationMixin):
                explanation_data["feature_weights"] = feature_weights
            if issubclass(self.return_element_type_, AtomExplanationMixin):
                explanation_data["atom_weights"] = atom_weights
            if issubclass(self.return_element_type_, BondExplanationMixin):
                explanation_data["bond_weights"] = bond_weights
            if issubclass(self.return_element_type_, SHAPExplanationMixin):
                explanation_data["expected_value"] = np.atleast_1d(
                    self.explainer.expected_value
                )

            explanation_results.append(self.return_element_type_(**explanation_data))

        return explanation_results


class SHAPTreeExplainer(SHAPExplainerAdapter):  # pylint: disable=too-few-public-methods
    """Wrapper for SHAP's TreeExplainer that can handle pipelines and molecules.

    Wraps SHAP's TreeExplainer to explain predictions of a pipeline containing a
    tree-based model.

    Note on failed instances:
    SHAPTreeExplainer will automatically handle fill values for failed instances and
    returns an invalid explanation for them. However, fill values that could be valid
    predictions, e.g. 0, are not necessarily detected. Set the fill value to np.nan or
    None if these failed instances should not be explained.
    """

    def __init__(
        self,
        pipeline: Pipeline,
        **kwargs: Any,
    ) -> None:
        """Initialize the SHAPKernelExplainer.

        Parameters
        ----------
        pipeline : Pipeline
            The pipeline containing the model to explain.
        kwargs : Any
            Additional keyword arguments for SHAP's Explainer.
        """
        explainer = self._create_explainer(pipeline, **kwargs)
        super().__init__(pipeline, explainer)

    @staticmethod
    def _create_explainer(pipeline: Pipeline, **kwargs: Any) -> shap.TreeExplainer:
        """Create the TreeExplainer object from shap.

        Parameters
        ----------
        pipeline : Pipeline
            The pipeline containing the model to explain.
        kwargs : Any
            Additional keyword arguments for the explainer.

        Returns
        -------
        shap.TreeExplainer
            The explainer object.
        """
        model = get_model_from_pipeline(pipeline, raise_not_found=True)
        explainer = shap.TreeExplainer(
            model,
            **kwargs,
        )
        return explainer


class SHAPKernelExplainer(
    SHAPExplainerAdapter
):  # pylint: disable=too-few-public-methods
    """Wrapper for SHAP's KernelExplainer that can handle pipelines and molecules."""

    def __init__(
        self,
        pipeline: Pipeline,
        **kwargs: Any,
    ) -> None:
        """Initialize the SHAPKernelExplainer.

        Parameters
        ----------
        pipeline : Pipeline
            The pipeline containing the model to explain.
        kwargs : Any
            Additional keyword arguments for SHAP's Explainer.
        """
        explainer = self._create_explainer(pipeline, **kwargs)
        super().__init__(pipeline, explainer)

    @staticmethod
    def _create_explainer(pipeline: Pipeline, **kwargs: Any) -> shap.KernelExplainer:
        """Create the explainer object.

        Parameters
        ----------
        pipeline : Pipeline
            The pipeline containing the model to explain.
        kwargs : Any
            Additional keyword arguments for the explainer.

        Returns
        -------
        shap.KernelExplainer
            The explainer object.
        """
        model = get_model_from_pipeline(pipeline, raise_not_found=True)
        prediction_function = _get_prediction_function(model)
        explainer = shap.KernelExplainer(
            prediction_function,
            **kwargs,
        )
        return explainer



"""Helper functions to extract subpipelines from a pipeline."""

from __future__ import annotations

from typing import Any, Callable

from sklearn.base import BaseEstimator

from molpipeline import FilterReinserter, Pipeline, PostPredictionWrapper
from molpipeline.abstract_pipeline_elements.core import (
    AnyToMolPipelineElement,
    MolToAnyPipelineElement,
)


def _get_molecule_reading_position_from_pipeline(pipeline: Pipeline) -> int | None:
    """Heuristic to select the position of the central molecule reading element in a pipeline.

    This function searches the last AnyToMolPipelineElement in the pipeline. We select the last
    AnyToMolPipelineElement in the pipeline because we have some standardization pipelines
    that write the molecule a smiles and read them back in to ensure they are readable.

    Parameters
    ----------
    pipeline: Pipeline
        The pipeline to search for the molecule reading element.

    Returns
    -------
    int | None
        The position of the molecule reading element in the pipeline.
    """
    for i, step in enumerate(reversed(pipeline.steps)):
        if isinstance(step[1], AnyToMolPipelineElement):
            return len(pipeline.steps) - i - 1
    return None


def _get_model_element_position_from_pipeline(pipeline: Pipeline) -> int | None:
    """Heuristic to select the position of the machine learning estimator model in a pipeline.

    The returned element should be the element returning the pipeline's predictions. So, the
    pipeline up to this element can be executed to obtain the predictions.

    Parameters
    ----------
    pipeline: Pipeline
        The pipeline to search for the model element.

    Returns
    -------
    int | None
        The position of the model element in the pipeline or None if no model element is found.
    """
    for i, step in enumerate(reversed(pipeline.steps)):
        if isinstance(step[1], BaseEstimator):
            if isinstance(step[1], PostPredictionWrapper):
                # skip PostPredictionWrappers.
                continue
            return len(pipeline.steps) - i - 1
    return None


def _get_featurization_element_position_from_pipeline(pipeline: Pipeline) -> int | None:
    """Heuristic to select the position of the featurization element in a pipeline.

    Parameters
    ----------
    pipeline: Pipeline
        The pipeline to search for the featurization element.

    Returns
    -------
    int | None
        The position of the featurization element in the pipeline or None if no featurization element is found.

    """
    for i, step in enumerate(reversed(pipeline.steps)):
        if isinstance(step[1], MolToAnyPipelineElement):
            return len(pipeline.steps) - i - 1
    return None


class SubpipelineExtractor:
    """A helper class to extract parts of a pipeline."""

    def __init__(self, pipeline: Pipeline) -> None:
        """Initialize the SubpipelineExtractor.

        Parameters
        ----------
        pipeline : Pipeline
            The pipeline to extract subpipelines from.
        """
        self.pipeline = pipeline

    def _get_index_of_element_by_id(self, element: Any) -> int | None:
        """Get the index of an element by id (the pointer or memory address).

        Parameters
        ----------
        element : Any
            The element to extract.

        Returns
        -------
        int | None
            The index of the element or None if the element was not found.
        """
        for i, (_, pipeline_element) in enumerate(self.pipeline.steps):
            if pipeline_element is element:
                return i
        return None

    def _get_index_of_element_by_name(self, element_name: str) -> int | None:
        """Get the index of an element by name.

        Parameters
        ----------
        element_name : str
            The name of the element to extract.

        Returns
        -------
        int | None
            The index of the element or None if the element was not found.
        """
        for i, (name, _) in enumerate(self.pipeline.steps):
            if name == element_name:
                return i
        return None

    def _extract_single_element_index(
        self,
        element_name: str | None,
        get_index_function: Callable[[Pipeline], int | None],
    ) -> int | None:
        """Extract the index of a single element from the pipeline.

        Parameters
        ----------
        element_name : str | None
            The name of the element to extract.
        get_index_function : Callable[[Pipeline], int | None]
            A function that returns the index of the element to extract.

        Returns
        -------
        Any | None
            The index of the extracted element or None if the element was not found.
        """
        if element_name is not None:
            return self._get_index_of_element_by_name(element_name)
        return get_index_function(self.pipeline)

    def _extract_single_element(
        self,
        element_name: str | None,
        get_index_function: Callable[[Pipeline], int | None],
    ) -> Any | None:
        """Extract a single element from the pipeline.

        Parameters
        ----------
        element_name : str | None
            The name of the element to extract.
        get_index_function : Callable[[Pipeline], int | None]
            A function that returns the index of the element to extract.

        Returns
        -------
        Any | None
            The extracted element or None if the element was not found.
        """
        if element_name is not None:
            # if a name is provided, access the element by name
            return self.pipeline.named_steps[element_name]
        element_index = self._extract_single_element_index(None, get_index_function)
        if element_index is None:
            return None
        return self.pipeline.steps[element_index][1]

    def get_molecule_reader_element(
        self, element_name: str | None = None
    ) -> AnyToMolPipelineElement | None:
        """Get the molecule reader element from the pipeline, e.g. a SmilesToMol element.

        Parameters
        ----------
        element_name : str | None
            The name of the element to extract.

        Returns
        -------
        AnyToMolPipelineElement | None
            The extracted molecule reader element or None if the element was not found.
        """
        return self._extract_single_element(
            element_name,
            _get_molecule_reading_position_from_pipeline,
        )

    def get_featurization_element(
        self, element_name: str | None = None
    ) -> BaseEstimator | None:
        """Get the featurization element from the pipeline, e.g., a MolToMorganFP element.

        Parameters
        ----------
        element_name : str | None
            The name of the element to extract.

        Returns
        -------
        BaseEstimator | None
            The extracted featurization element or None if the element was not found.
        """
        return self._extract_single_element(
            element_name, _get_featurization_element_position_from_pipeline
        )

    def get_model_element(
        self, element_name: str | None = None
    ) -> BaseEstimator | None:
        """Get the machine learning model element from the pipeline, e.g. a RandomForestClassifier.

        Parameters
        ----------
        element_name : str | None
            The name of the element to extract.

        Returns
        -------
        BaseEstimator | None
            The extracted model element or None if the element was not found.
        """
        return self._extract_single_element(
            element_name, _get_model_element_position_from_pipeline
        )

    def _get_subpipline_from_start(
        self,
        element_name: str | None,
        start_get_index_function: Callable[[Pipeline], int | None],
    ) -> Pipeline | None:
        """Get a subpipeline up to a specific element starting from the first element of the original pipeline.

        Parameters
        ----------
        element_name : str | None
            The name of the element to extract.
        start_get_index_function : Callable[[Pipeline], int | None]
            A function that returns the index of the subpipline's last element.

        Returns
        -------
        Pipeline | None
            The extracted subpipeline or None if the corresponding last element was not found.
        """
        element_index = self._extract_single_element_index(
            element_name, start_get_index_function
        )
        if element_index is None:
            return None
        return Pipeline(steps=self.pipeline.steps[: element_index + 1])

    def get_molecule_reader_subpipeline(
        self, element_name: str | None = None
    ) -> Pipeline | None:
        """Get a subpipeline up to the molecule reading element.

        Note that standardization steps, like salt removal, are not guaranteed to be included.

        Parameters
        ----------
        element_name : str | None
            The name of the last element in the subpipeline to extract.

        Returns
        -------
        Pipeline | None
            The extracted subpipeline or None if the corresponding last element was not found.
        """
        return self._get_subpipline_from_start(
            element_name, _get_molecule_reading_position_from_pipeline
        )

    def get_featurization_subpipeline(
        self, element_name: str | None = None
    ) -> Pipeline | None:
        """Get a subpipeline up to the featurization element.

        Parameters
        ----------
        element_name : str | None
            The name of the last element in the subpipeline to extract.

        Returns
        -------
        Pipeline | None
            The extracted subpipeline or None if the corresponding last element was not found.
        """
        return self._get_subpipline_from_start(
            element_name, _get_featurization_element_position_from_pipeline
        )

    def get_model_subpipeline(self, element_name: str | None = None) -> Pipeline | None:
        """Get a subpipeline up to the machine learning model element.

        Parameters
        ----------
        element_name : str | None
            The name of the last element in the subpipeline to extract.

        Returns
        -------
        Pipeline | None
            The extracted subpipeline or None if the corresponding last element was not found.
        """
        return self._get_subpipline_from_start(
            element_name, _get_model_element_position_from_pipeline
        )

    def get_subpipeline(
        self,
        first_element: Any,
        second_element: Any,
        first_offset: int = 0,
        second_offset: int = 0,
    ) -> Pipeline | None:
        """Get a subpipeline between two elements.

        This function only checks the names of the elements.
        If the elements are not found or the second element is before the first element, a ValueError is raised.

        Parameters
        ----------
        first_element : Any
            The first element of the subpipeline.
        second_element : Any
            The second element of the subpipeline.
        first_offset : int
            The offset to apply to the first element.
        second_offset : int
            The offset to apply to the second element.

        Returns
        -------
        Pipeline | None
            The extracted subpipeline or None if the elements were not found.
        """
        first_element_index = self._get_index_of_element_by_id(first_element)
        if first_element_index is None:
            raise ValueError(f"Element {first_element} not found in pipeline.")
        second_element_index = self._get_index_of_element_by_id(second_element)
        if second_element_index is None:
            raise ValueError(f"Element {second_element} not found in pipeline.")

        # apply user-defined offsets
        first_element_index += first_offset
        second_element_index += second_offset

        if second_element_index < first_element_index:
            raise ValueError(
                f"Element {second_element} must be after element {first_element}."
            )
        return Pipeline(
            steps=self.pipeline.steps[first_element_index : second_element_index + 1]
        )

    def get_all_filter_reinserter_fill_values(self) -> list[Any]:
        """Get all fill values for FilterReinserter elements in the pipeline.

        Returns
        -------
        list[Any]
            The fill values for all FilterReinserter elements in the pipeline.
        """
        fill_values = set()
        for _, step in self.pipeline.steps:
            if isinstance(step, FilterReinserter):
                fill_values.add(step.fill_value)
            if isinstance(step, PostPredictionWrapper) and isinstance(
                step.wrapped_estimator, FilterReinserter
            ):
                fill_values.add(step.wrapped_estimator.fill_value)
        return list(fill_values)


def get_featurization_subpipeline(
    pipeline: Pipeline, raise_not_found: bool = False
) -> Pipeline | None:
    """Get the featurization subpipeline from a pipeline.

    Parameters
    ----------
    pipeline : Pipeline
        The pipeline to extract the featurization subpipeline from.
    raise_not_found : bool
        If True, raise a ValueError if the model was not found.

    Raises
    ------
    ValueError
        If the model was not found and raise_not_found is True.

    Returns
    -------
    Pipeline | None
        The extracted featurization subpipeline or None if the featurization element was not found.
    """
    pipeline_extractor = SubpipelineExtractor(pipeline)
    featurization_subpipeline = pipeline_extractor.get_featurization_subpipeline()
    if raise_not_found and featurization_subpipeline is None:
        raise ValueError("Could not determine the featurization subpipeline.")
    return featurization_subpipeline


def get_model_from_pipeline(
    pipeline: Pipeline, raise_not_found: bool = False
) -> BaseEstimator | None:
    """Get the model from a pipeline.

    Parameters
    ----------
    pipeline : Pipeline
        The pipeline to extract the model from.
    raise_not_found : bool
        If True, raise a ValueError if the model was not found.

    Raises
    ------
    ValueError
        If the model was not found and raise_not_found is True.

    Returns
    -------
    BaseEstimator | None
        The extracted model or None if the model was not found.
    """
    pipeline_extractor = SubpipelineExtractor(pipeline)
    model = pipeline_extractor.get_model_element()
    if raise_not_found and model is None:
        raise ValueError("Could not determine the model to explain.")
    return model



"""Sklearn estimators for computing similarity and distance matrices."""

from __future__ import annotations

from typing import Any

try:
    from typing import Self  # pylint: disable=no-name-in-module
except ImportError:
    from typing_extensions import Self

import numpy as np
import numpy.typing as npt
from scipy.sparse import csr_matrix
from sklearn.base import BaseEstimator, TransformerMixin

from molpipeline.utils.kernel import tanimoto_similarity_sparse


class TanimotoToTraining(BaseEstimator, TransformerMixin):
    """Transformer for computing tanimoto similarity matrices to data seen during training.

    Can also be used to compute distance matrices.

    Attributes
    ----------
    training_matrix: npt.NDArray[np.float64] | csr_matrix | None
        Features seen during fit.
    """

    training_matrix: npt.NDArray[np.float64] | csr_matrix | None

    def __init__(self, distance: bool = False) -> None:
        """Initialize TanimotoSimilarityToTraining.

        Parameters
        ----------
        distance: bool, optional
            If True, the distance matrix is computed, by default False
            The distance matrix is computed as 1 - similarity_matrix.
        """
        self.training_matrix = None
        self.distance = distance

    def _sim(
        self,
        matrix_a: npt.NDArray[np.float64] | csr_matrix,
        matrix_b: npt.NDArray[np.float64] | csr_matrix,
    ) -> npt.NDArray[np.float64]:
        """Compute the similarity matrix.

        Parameters
        ----------
        matrix_a : npt.NDArray[np.float64] | csr_matrix
            First matrix.
        matrix_b : npt.NDArray[np.float64] | csr_matrix
            Second matrix.

        Returns
        -------
        npt.NDArray[np.float64]
            Similarity matrix. If distance is True, the distance matrix is computed instead.
        """
        if not isinstance(matrix_a, csr_matrix):
            matrix_a = csr_matrix(matrix_a)
        if not isinstance(matrix_b, csr_matrix):
            matrix_b = csr_matrix(matrix_b)
        if self.distance:
            return 1 - tanimoto_similarity_sparse(matrix_a, matrix_b)  # type: ignore
        return tanimoto_similarity_sparse(matrix_a, matrix_b)

    def fit(
        self,
        X: npt.NDArray[np.float64] | csr_matrix,  # pylint: disable=invalid-name
        y: npt.NDArray[np.float64] | None = None,  # pylint: disable=unused-argument
    ) -> Self:
        """Fit the model.

        Parameters
        ----------
        X : npt.NDArray[np.float64] | csr_matrix
            Feature matrix to which the similarity matrix is computed.
        y : npt.NDArray[np.float64] | None, optional
            Labels, by default None and never used

        Returns
        -------
        Self
            Fitted model.
        """
        self.training_matrix = X
        return self

    def transform(
        self, X: npt.NDArray[np.float64] | csr_matrix  # pylint: disable=invalid-name
    ) -> npt.NDArray[np.float64]:
        """Transform the data.

        Parameters
        ----------
        X : npt.NDArray[np.float64] | csr_matrix
            Feature matrix to which the similarity matrix is computed.

        Returns
        -------
        npt.NDArray[np.float64]
            Similarity matrix of X to the training matrix.
        """
        if self.training_matrix is None:
            raise ValueError("Please fit the transformer before transforming!")
        return self._sim(X, self.training_matrix)

    def fit_transform(
        self,
        X: npt.NDArray[np.float64] | csr_matrix,  # pylint: disable=invalid-name
        y: npt.NDArray[np.float64] | None = None,
        **fit_params: Any,
    ) -> npt.NDArray[np.float64]:
        """Fit the model and transform the data.

        Parameters
        ----------
        X: npt.NDArray[np.float64] | csr_matrix
            Feature matrix to fit the model. Is returned as similarity matrix to itself.
        y: npt.NDArray[np.float64] | None, optional
            Labels, by default None and never used
        **fit_params: Any
            Additional fit parameters. Ignored.

        Returns
        -------
        npt.NDArray[np.float64]
            Similarity matrix of X to itself.
        """
        self.fit(X, y)
        return self.transform(X)



"""Unittests for testing conversion of various inputs to RDKit molecules."""

import gzip
import unittest

from rdkit import Chem, rdBase

from molpipeline import Pipeline
from molpipeline.abstract_pipeline_elements.core import InvalidInstance
from molpipeline.any2mol import AutoToMol, BinaryToMol, SDFToMol, SmilesToMol
from tests import TEST_DATA_DIR

# pylint: disable=duplicate-code  # test case molecules are allowed to be duplicated
SMILES_ANTIMONY = "[SbH6+3]"
SMILES_BENZENE = "c1ccccc1"
SMILES_CHLOROBENZENE = "Clc1ccccc1"
SMILES_CL_BR = "NC(Cl)(Br)C(=O)O"
SMILES_METAL_AU = "OC[C@H]1OC(S[Au])[C@H](O)[C@@H](O)[C@@H]1O"

INCHI_BENZENE = "InChI=1S/C6H6/c1-2-4-6-5-3-1/h1-6H"
INCHI_CHLOROBENZENE = "InChI=1S/C6H5Cl/c7-6-4-2-1-3-5-6/h1-5H"

# SDF
with gzip.open(TEST_DATA_DIR / "P86_B_400.sdf.gz") as file:
    SDF_P86_B_400 = file.read().decode("UTF-8")
_sdf_supplier = Chem.SDMolSupplier()
_sdf_supplier.SetData(SDF_P86_B_400)

# INCHI
INCHI_P86_LIGAND = "InChI=1S/C16H24I2N2O/c1-3-20(4-2)14-5-7-19(8-6-14)11-12-9-13(17)10-15(18)16(12)21/h9-10,14,21H,3-8,11H2,1-2H3"


# RDKit mols
MOL_ANTIMONY = Chem.MolFromSmiles(SMILES_ANTIMONY)
MOL_BENZENE = Chem.MolFromSmiles(SMILES_BENZENE)
MOL_CHLOROBENZENE = Chem.MolFromSmiles(SMILES_CHLOROBENZENE)
MOL_CL_BR = Chem.MolFromSmiles(SMILES_CL_BR)
MOL_METAL_AU = Chem.MolFromSmiles(SMILES_METAL_AU)
MOL_P86_LIGAND = next(_sdf_supplier)


class TestAuto2Mol(unittest.TestCase):
    """Test case for testing conversion of input to molecules."""

    def test_auto2mol_for_smiles(self) -> None:
        """Test molecules can be read from smiles automatically."""

        test_smiles = [
            SMILES_ANTIMONY,
            SMILES_BENZENE,
            SMILES_CHLOROBENZENE,
            SMILES_CL_BR,
            SMILES_METAL_AU,
        ]
        expected_mols = [
            MOL_ANTIMONY,
            MOL_BENZENE,
            MOL_CHLOROBENZENE,
            MOL_CL_BR,
            MOL_METAL_AU,
        ]

        pipeline = Pipeline(
            [
                (
                    "Auto2Mol",
                    AutoToMol(
                        elements=(
                            SmilesToMol(),
                            BinaryToMol(),
                            SDFToMol(),
                        )
                    ),
                ),
            ]
        )
        log_block = rdBase.BlockLogs()
        actual_mols = pipeline.fit_transform(test_smiles)
        self.assertEqual(len(test_smiles), len(actual_mols))
        self.assertTrue(
            all(
                Chem.MolToInchi(smiles_mol) == Chem.MolToInchi(original_mol)
                for smiles_mol, original_mol in zip(actual_mols, expected_mols)
            )
        )
        del log_block

    def test_auto2mol_for_inchi(self) -> None:
        """Test molecules can be read from inchi automatically."""

        test_inchis = [INCHI_BENZENE, INCHI_CHLOROBENZENE]
        expected_mols = [MOL_BENZENE, MOL_CHLOROBENZENE]

        pipeline = Pipeline(
            [
                (
                    "Auto2Mol",
                    AutoToMol(),
                ),
            ]
        )
        log_block = rdBase.BlockLogs()
        actual_mols = pipeline.fit_transform(test_inchis)
        self.assertEqual(len(test_inchis), len(actual_mols))
        self.assertTrue(
            all(
                Chem.MolToInchi(smiles_mol) == Chem.MolToInchi(original_mol)
                for smiles_mol, original_mol in zip(actual_mols, expected_mols)
            )
        )
        del log_block

    def test_auto2mol_for_sdf(self) -> None:
        """Test molecules can be read from sdf automatically."""

        pipeline = Pipeline(
            [
                (
                    "Auto2Mol",
                    AutoToMol(
                        elements=(
                            SmilesToMol(),
                            BinaryToMol(),
                            SDFToMol(),
                        )
                    ),
                ),
            ]
        )
        log_block = rdBase.BlockLogs()
        actual_mols = pipeline.fit_transform([SDF_P86_B_400])
        self.assertEqual(len(actual_mols), 1)
        self.assertTrue(
            Chem.MolToInchi(actual_mols[0]).startswith(INCHI_P86_LIGAND),
        )
        del log_block

    def test_auto2mol_for_binary(self) -> None:
        """Test molecules can be read from binary automatically."""

        test_bin_mols = [
            MOL_ANTIMONY.ToBinary(),
            MOL_BENZENE.ToBinary(),
            MOL_CHLOROBENZENE.ToBinary(),
            MOL_CL_BR.ToBinary(),
            MOL_METAL_AU.ToBinary(),
        ]
        expected_mols = [
            MOL_ANTIMONY,
            MOL_BENZENE,
            MOL_CHLOROBENZENE,
            MOL_CL_BR,
            MOL_METAL_AU,
        ]

        pipeline = Pipeline(
            [
                (
                    "Auto2Mol",
                    AutoToMol(
                        elements=(
                            SmilesToMol(),
                            BinaryToMol(),
                            SDFToMol(),
                        )
                    ),
                ),
            ]
        )
        log_block = rdBase.BlockLogs()
        actual_mols = pipeline.fit_transform(test_bin_mols)
        self.assertEqual(len(test_bin_mols), len(actual_mols))
        self.assertTrue(
            all(
                Chem.MolToInchi(smiles_mol) == Chem.MolToInchi(original_mol)
                for smiles_mol, original_mol in zip(actual_mols, expected_mols)
            )
        )
        del log_block

    def test_auto2mol_for_molecule(self) -> None:
        """Test molecules can be read from molecules automatically."""

        test_mols = [
            MOL_ANTIMONY,
            MOL_BENZENE,
            MOL_CHLOROBENZENE,
            MOL_CL_BR,
            MOL_METAL_AU,
        ]
        expected_mols = [
            MOL_ANTIMONY,
            MOL_BENZENE,
            MOL_CHLOROBENZENE,
            MOL_CL_BR,
            MOL_METAL_AU,
        ]

        pipeline = Pipeline(
            [
                (
                    "Auto2Mol",
                    AutoToMol(
                        elements=(
                            SmilesToMol(),
                            BinaryToMol(),
                            SDFToMol(),
                        )
                    ),
                ),
            ]
        )
        log_block = rdBase.BlockLogs()
        actual_mols = pipeline.fit_transform(test_mols)
        self.assertEqual(len(test_mols), len(actual_mols))
        self.assertTrue(
            all(
                Chem.MolToInchi(smiles_mol) == Chem.MolToInchi(original_mol)
                for smiles_mol, original_mol in zip(actual_mols, expected_mols)
            )
        )
        del log_block

    def test_auto2mol_mixed_inputs(self) -> None:
        """Test molecules can be read from mixed inputs automatically."""

        test_inputs = [
            SDF_P86_B_400,
            MOL_ANTIMONY.ToBinary(),
            MOL_BENZENE,
            SMILES_CHLOROBENZENE,
            MOL_CL_BR.ToBinary(),
            SMILES_METAL_AU,
        ]
        test_mols = [
            MOL_P86_LIGAND,
            MOL_ANTIMONY,
            MOL_BENZENE,
            MOL_CHLOROBENZENE,
            MOL_CL_BR,
            MOL_METAL_AU,
        ]

        pipeline = Pipeline(
            [
                (
                    "Auto2Mol",
                    AutoToMol(
                        elements=(
                            SmilesToMol(),
                            BinaryToMol(),
                            SDFToMol(),
                        )
                    ),
                ),
            ]
        )
        log_block = rdBase.BlockLogs()
        actual_mols = pipeline.fit_transform(test_inputs)
        self.assertEqual(len(test_inputs), len(actual_mols))
        self.assertTrue(
            all(
                Chem.MolToInchi(smiles_mol) == Chem.MolToInchi(original_mol)
                for smiles_mol, original_mol in zip(actual_mols, test_mols)
            )
        )
        del log_block

    def test_auto2mol_invalid_input_nones(self) -> None:
        """Test molecules can be read from invalid input with Nones automatically."""

        test_inputs = [
            SDF_P86_B_400,
            None,
            MOL_BENZENE,
            None,
        ]

        pipeline = Pipeline(
            [
                (
                    "Auto2Mol",
                    AutoToMol(
                        elements=(
                            SmilesToMol(),
                            BinaryToMol(),
                            SDFToMol(),
                        )
                    ),
                ),
            ]
        )
        log_block = rdBase.BlockLogs()
        actual_mols = pipeline.fit_transform(test_inputs)
        self.assertEqual(len(test_inputs), len(actual_mols))

        self.assertEqual(
            Chem.MolToInchi(actual_mols[0]), Chem.MolToInchi(MOL_P86_LIGAND)
        )
        self.assertTrue(isinstance(actual_mols[1], InvalidInstance))
        self.assertEqual(Chem.MolToInchi(actual_mols[2]), Chem.MolToInchi(MOL_BENZENE))
        self.assertTrue(isinstance(actual_mols[3], InvalidInstance))
        del log_block

    def test_auto2mol_invalid_input_no_matching_reader(self) -> None:
        """Test molecules can be read from invalid input with no matching readers automatically."""

        test_inputs = [
            SDF_P86_B_400,
            MOL_BENZENE,
            MOL_BENZENE.ToBinary(),
            SMILES_CHLOROBENZENE,
        ]

        pipeline = Pipeline(
            [
                (
                    "Auto2Mol",
                    AutoToMol(elements=(SmilesToMol(),)),
                ),
            ]
        )
        log_block = rdBase.BlockLogs()
        actual_mols = pipeline.fit_transform(test_inputs)
        self.assertEqual(len(test_inputs), len(actual_mols))

        self.assertTrue(isinstance(actual_mols[0], InvalidInstance))
        self.assertEqual(Chem.MolToInchi(actual_mols[1]), Chem.MolToInchi(MOL_BENZENE))
        self.assertTrue(isinstance(actual_mols[2], InvalidInstance))
        self.assertEqual(
            Chem.MolToInchi(actual_mols[3]), Chem.MolToInchi(MOL_CHLOROBENZENE)
        )
        del log_block

    def test_auto2mol_invalid_input_empty_elements(self) -> None:
        """Test molecules can be read from invalid input with no reader elements automatically."""

        test_inputs = [
            SDF_P86_B_400,
            MOL_BENZENE,
            MOL_BENZENE.ToBinary(),
            SMILES_CHLOROBENZENE,
        ]

        pipeline = Pipeline(
            [
                (
                    "Auto2Mol",
                    AutoToMol(elements=()),
                ),
            ]
        )
        log_block = rdBase.BlockLogs()
        actual_mols = pipeline.fit_transform(test_inputs)
        self.assertEqual(len(test_inputs), len(actual_mols))

        self.assertTrue(isinstance(actual_mols[0], InvalidInstance))
        self.assertEqual(Chem.MolToInchi(actual_mols[1]), Chem.MolToInchi(MOL_BENZENE))
        self.assertTrue(isinstance(actual_mols[2], InvalidInstance))
        self.assertTrue(isinstance(actual_mols[3], InvalidInstance))
        del log_block



"""MolToNetCharge pipeline element."""

from __future__ import annotations

import copy
from typing import Any, Literal, TypeAlias

try:
    from typing import Self  # type: ignore[attr-defined]
except ImportError:
    from typing_extensions import Self

import numpy as np
import numpy.typing as npt
from rdkit import Chem
from sklearn.preprocessing import StandardScaler

from molpipeline.abstract_pipeline_elements.core import InvalidInstance
from molpipeline.abstract_pipeline_elements.mol2any.mol2floatvector import (
    MolToDescriptorPipelineElement,
)
from molpipeline.utils.molpipeline_types import AnyTransformer, RDKitMol

# Methods to compute the net charge of a molecule.
# - "formal_charge" uses the formal charges of the atoms with rdkit.Chem.rdmolops.GetFormalCharge
# - "gasteiger" uses the Gasteiger charges of the atoms with rdkit.Chem.rdPartialCharges.ComputeGasteigerCharges
MolToNetChargeMethod: TypeAlias = Literal["formal_charge", "gasteiger"]


class MolToNetCharge(MolToDescriptorPipelineElement):
    """PipelineElement for calculating molecules' net charge."""

    def __init__(
        self,
        charge_method: MolToNetChargeMethod = "formal_charge",
        standardizer: AnyTransformer | None = StandardScaler(),
        name: str = "MolToNetCharge",
        n_jobs: int = 1,
        uuid: str | None = None,
    ) -> None:
        """Initialize MolToNetCharge.

        Parameters
        ----------
        charge_method: MolToNetChargeMethod, optional (default="formal_charge")
            Policy how to compute the net charge of a molecule. Can be "formal_charge" which uses sum
            of the formal charges assigned to each atom. "gasteiger" computes the Gasteiger partial
            charges and returns the rounded sum over the atoms.
        standardizer: AnyTransformer, optional
            Standardizer to use, by default StandardScaler()
        name: str, optional
            Name of the pipeline element, by default "MolToNetCharge"
        n_jobs: int, optional
            Number of jobs to run in parallel, by default 1
        uuid: str, optional
            UUID of the pipeline element, by default None
        """
        self._descriptor_list = ["NetCharge"]
        self._feature_names = self._descriptor_list
        self._charge_method = charge_method
        # pylint: disable=R0801
        super().__init__(
            standardizer=standardizer,
            name=name,
            n_jobs=n_jobs,
            uuid=uuid,
        )

    @property
    def n_features(self) -> int:
        """Return the number of features."""
        return len(self._descriptor_list)

    @property
    def descriptor_list(self) -> list[str]:
        """Return a copy of the descriptor list. Alias of `feature_names`."""
        return self._descriptor_list[:]

    def _get_net_charge_gasteiger(
        self, value: RDKitMol
    ) -> npt.NDArray[np.float64] | InvalidInstance:
        """Transform a single molecule to it's net charge using Gasteiger charges.

        Based on https://github.com/rdkit/rdkit/discussions/4331

        Parameters
        ----------
        value: RDKitMol
            RDKit molecule to transform.

        Returns
        -------
        Optional[npt.NDArray[np.float64]]
            Net charge of the given molecule.
        """
        # copy molecule since ComputeGasteigerCharges modifies the molecule inplace
        value_copy = Chem.Mol(value)
        Chem.rdPartialCharges.ComputeGasteigerCharges(value_copy)
        atoms_contributions = np.array(
            [atom.GetDoubleProp("_GasteigerCharge") for atom in value_copy.GetAtoms()]
        )
        if np.any(np.isnan(atoms_contributions)):
            return InvalidInstance(self.uuid, "NaN in Gasteiger charges", self.name)
        # sum up the charges and round to the nearest integer.
        net_charge = np.round(np.sum(atoms_contributions, keepdims=True))
        return net_charge

    def pretransform_single(
        self, value: RDKitMol
    ) -> npt.NDArray[np.float64] | InvalidInstance:
        """Transform a single molecule to it's net charge.

        Parameters
        ----------
        value: RDKitMol
            RDKit molecule to transform.

        Returns
        -------
        Optional[npt.NDArray[np.float64]]
            Net charge of the given molecule.
        """
        if self._charge_method == "formal_charge":
            return np.array([Chem.GetFormalCharge(value)], dtype=np.float64)
        if self._charge_method == "gasteiger":
            return self._get_net_charge_gasteiger(value)
        raise ValueError(f"Unknown charge policy: {self._charge_method}")

    def get_params(self, deep: bool = True) -> dict[str, Any]:
        """Get the parameters of the pipeline element.

        Parameters
        ----------
        deep: bool
            If true create a deep copy of the parameters

        Returns
        -------
        dict[str, Any]
            Parameter of the pipeline element.
        """
        parent_dict = dict(super().get_params(deep=deep))
        if deep:
            parent_dict["charge_policy"] = copy.deepcopy(self._charge_method)
        else:
            parent_dict["charge_policy"] = self._charge_method
        return parent_dict

    def set_params(self, **parameters: Any) -> Self:
        """Set parameters.

        Parameters
        ----------
        parameters: Any
            Parameters to set

        Returns
        -------
        Self
            Self
        """
        parameters_shallow_copy = dict(parameters)
        charge_policy = parameters_shallow_copy.pop("charge_policy", None)
        if charge_policy is not None:
            self._charge_method = charge_policy
        super().set_params(**parameters_shallow_copy)
        return self



"""Classes ment to transform given input to a RDKit molecule."""

from __future__ import annotations

from typing import Any, Optional

try:
    from typing import Self  # type: ignore[attr-defined]
except ImportError:
    from typing_extensions import Self

from rdkit import Chem

from molpipeline.abstract_pipeline_elements.any2mol.string2mol import (
    SimpleStringToMolElement,
)
from molpipeline.utils.molpipeline_types import RDKitMol


class SmilesToMol(SimpleStringToMolElement):
    """Transforms Smiles to RDKit Mol objects."""

    def __init__(
        self,
        remove_hydrogens: bool = True,
        name: str = "smiles2mol",
        n_jobs: int = 1,
        uuid: Optional[str] = None,
    ) -> None:
        """Initialize SmilesToMol object.

        Parameters
        ----------
        remove_hydrogens: bool
            Whether to remove hydrogens from the molecule.
        name: str
            Name of the object.
        n_jobs: int
            Number of jobs to run in parallel.
        uuid: Optional[str]
            UUID of the object.
        """
        super().__init__(name=name, n_jobs=n_jobs, uuid=uuid)
        self._remove_hydrogens = remove_hydrogens

    def _get_parser_config(self) -> Chem.SmilesParserParams:
        """Get parser configuration.

        Returns
        -------
        dict[str, Any]
            Configuration for the parser.
        """
        # set up rdkit smiles parser parameters
        parser_params = Chem.SmilesParserParams()
        parser_params.removeHs = self._remove_hydrogens
        return parser_params

    def string_to_mol(self, value: str) -> RDKitMol:
        """Transform Smiles string to molecule.

        Parameters
        ----------
        value: str
            SMILES string.

        Returns
        -------
        RDKitMol
            Rdkit molecule if valid SMILES, else None.
        """
        return Chem.MolFromSmiles(value, self._get_parser_config())

    def get_params(self, deep: bool = True) -> dict[str, Any]:
        """Get parameters for this object.

        Parameters
        ----------
        deep: bool
            If True, return a deep copy of the parameters.

        Returns
        -------
        dict[str, Any]
            Dictionary of parameters.
        """
        parameters = super().get_params(deep)
        if deep:
            parameters["remove_hydrogens"] = bool(self._remove_hydrogens)

        else:
            parameters["remove_hydrogens"] = self._remove_hydrogens
        return parameters

    def set_params(self, **parameters: Any) -> Self:
        """Set parameters.

        Parameters
        ----------
        parameters: Any
            Dictionary of parameter names and values.

        Returns
        -------
        Self
            SmilesToMol pipeline element with updated parameters.
        """
        parameter_copy = dict(parameters)
        remove_hydrogens = parameter_copy.pop("remove_hydrogens", None)
        if remove_hydrogens is not None:
            self._remove_hydrogens = remove_hydrogens
        super().set_params(**parameter_copy)
        return self



"""Classes to filter molecule lists."""

from __future__ import annotations

from collections import Counter
from typing import Any, Mapping, Optional, Sequence, Union

try:
    from typing import Self  # type: ignore[attr-defined]
except ImportError:
    from typing_extensions import Self

from loguru import logger
from rdkit import Chem
from rdkit.Chem import Descriptors

from molpipeline.abstract_pipeline_elements.core import (
    InvalidInstance,
)
from molpipeline.abstract_pipeline_elements.core import (
    MolToMolPipelineElement as _MolToMolPipelineElement,
)
from molpipeline.abstract_pipeline_elements.mol2mol import (
    BaseKeepMatchesFilter as _BaseKeepMatchesFilter,
)
from molpipeline.abstract_pipeline_elements.mol2mol import (
    BasePatternsFilter as _BasePatternsFilter,
)
from molpipeline.abstract_pipeline_elements.mol2mol.filter import (
    FilterModeType,
    _within_boundaries,
)
from molpipeline.utils.molpipeline_types import (
    FloatCountRange,
    IntCountRange,
    IntOrIntCountRange,
    OptionalMol,
    RDKitMol,
)
from molpipeline.utils.value_conversions import count_value_to_tuple


class ElementFilter(_MolToMolPipelineElement):
    """ElementFilter which removes molecules containing chemical elements other than specified.

    Molecular elements are filtered based on their atomic number.
    The filter can be configured to allow only specific elements and/or a specific number of atoms of each element.
    """

    DEFAULT_ALLOWED_ELEMENT_NUMBERS = [
        1,
        5,
        6,
        7,
        8,
        9,
        14,
        15,
        16,
        17,
        34,
        35,
        53,
    ]

    def __init__(
        self,
        allowed_element_numbers: Optional[
            Union[list[int], dict[int, IntOrIntCountRange]]
        ] = None,
        add_hydrogens: bool = True,
        name: str = "ElementFilter",
        n_jobs: int = 1,
        uuid: Optional[str] = None,
    ) -> None:
        """Initialize ElementFilter.

        Parameters
        ----------
        allowed_element_numbers: Optional[Union[list[int], dict[int, IntOrIntCountRange]]]
            List of atomic numbers of elements to allowed in molecules. Per default allowed elements are:
            H, B, C, N, O, F, Si, P, S, Cl, Se, Br, I.
            Alternatively, a dictionary can be passed with atomic numbers as keys and an int for exact count or a tuple of minimum and maximum
        add_hydrogens: bool, optional (default: True)
            If True, in case Hydrogens are in allowed_element_list, add hydrogens to the molecule before filtering.
        name: str, optional (default: "ElementFilterPipe")
            Name of the pipeline element.
        n_jobs: int, optional (default: 1)
            Number of parallel jobs to use.
        uuid: str, optional (default: None)
            Unique identifier of the pipeline element.
        """
        super().__init__(name=name, n_jobs=n_jobs, uuid=uuid)
        self.allowed_element_numbers = allowed_element_numbers  # type: ignore
        self.add_hydrogens = add_hydrogens

    @property
    def add_hydrogens(self) -> bool:
        """Get add_hydrogens."""
        return self._add_hydrogens

    @add_hydrogens.setter
    def add_hydrogens(self, add_hydrogens: bool) -> None:
        """Set add_hydrogens.

        Parameters
        ----------
        add_hydrogens: bool
            If True, in case Hydrogens are in allowed_element_list, add hydrogens to the molecule before filtering.
        """
        self._add_hydrogens = add_hydrogens
        if self.add_hydrogens and 1 in self.allowed_element_numbers:
            self.process_hydrogens = True
        else:
            if 1 in self.allowed_element_numbers:
                logger.warning(
                    "Hydrogens are included in allowed_element_numbers, but add_hydrogens is set to False. "
                    "Thus hydrogens are NOT added before filtering. You might receive unexpected results."
                )
            self.process_hydrogens = False

    @property
    def allowed_element_numbers(self) -> dict[int, IntCountRange]:
        """Get allowed element numbers as dict."""
        return self._allowed_element_numbers

    @allowed_element_numbers.setter
    def allowed_element_numbers(
        self,
        allowed_element_numbers: Optional[
            Union[list[int], dict[int, IntOrIntCountRange]]
        ],
    ) -> None:
        """Set allowed element numbers as dict.

        Parameters
        ----------
        allowed_element_numbers: Optional[Union[list[int], dict[int, IntOrIntCountRange]]
            List of atomic numbers of elements to allowed in molecules.
        """
        self._allowed_element_numbers: dict[int, IntCountRange]
        if allowed_element_numbers is None:
            allowed_element_numbers = self.DEFAULT_ALLOWED_ELEMENT_NUMBERS
        if isinstance(allowed_element_numbers, (list, set)):
            self._allowed_element_numbers = {
                atom_number: (0, None) for atom_number in allowed_element_numbers
            }
        else:
            self._allowed_element_numbers = {
                int(atom_number): count_value_to_tuple(count)
                for atom_number, count in allowed_element_numbers.items()
            }

    def get_params(self, deep: bool = True) -> dict[str, Any]:
        """Get parameters of ElementFilter.

        Parameters
        ----------
        deep: bool, optional (default: True)
            If True, return the parameters of all subobjects that are PipelineElements.

        Returns
        -------
        dict[str, Any]
            Parameters of ElementFilter.
        """
        params = super().get_params(deep=deep)
        if deep:
            params["allowed_element_numbers"] = {
                atom_number: (count_tuple[0], count_tuple[1])
                for atom_number, count_tuple in self.allowed_element_numbers.items()
            }
        else:
            params["allowed_element_numbers"] = self.allowed_element_numbers
        params["add_hydrogens"] = self.add_hydrogens
        return params

    def set_params(self, **parameters: Any) -> Self:
        """Set parameters of ElementFilter.

        Parameters
        ----------
        parameters: Any
            Parameters to set.

        Returns
        -------
        Self
            Self.
        """
        parameter_copy = dict(parameters)
        if "allowed_element_numbers" in parameter_copy:
            self.allowed_element_numbers = parameter_copy.pop("allowed_element_numbers")
        if "add_hydrogens" in parameter_copy:
            self.add_hydrogens = parameter_copy.pop("add_hydrogens")
        super().set_params(**parameter_copy)
        return self

    def pretransform_single(self, value: RDKitMol) -> OptionalMol:
        """Invalidate molecule containing chemical elements that are not allowed.

        Parameters
        ----------
        value: RDKitMol
            Molecule to check.

        Returns
        -------
        OptionalMol
            Molecule if it contains only allowed elements, else InvalidInstance.
        """
        to_process_value = Chem.AddHs(value) if self.process_hydrogens else value
        elements_list = [atom.GetAtomicNum() for atom in to_process_value.GetAtoms()]
        elements_counter = Counter(elements_list)
        if any(
            element not in self.allowed_element_numbers for element in elements_counter
        ):
            return InvalidInstance(
                self.uuid, "Molecule contains forbidden chemical element.", self.name
            )
        for element, (lower_limit, upper_limit) in self.allowed_element_numbers.items():
            count = elements_counter[element]
            if not _within_boundaries(lower_limit, upper_limit, count):
                return InvalidInstance(
                    self.uuid,
                    f"Molecule contains forbidden number of element {element}.",
                    self.name,
                )
        return value


class SmartsFilter(_BasePatternsFilter):
    """Filter to keep or remove molecules based on SMARTS patterns.

    Notes
    -----
    There are four possible scenarios:
        - mode = "any" & keep_matches = True: Needs to match at least one filter element.
        - mode = "any" & keep_matches = False: Must not match any filter element.
        - mode = "all" & keep_matches = True: Needs to match all filter elements.
        - mode = "all" & keep_matches = False: Must not match all filter elements.
    """

    def _pattern_to_mol(self, pattern: str) -> RDKitMol:
        """Convert SMARTS pattern to RDKit molecule.

        Parameters
        ----------
        pattern: str
            SMARTS pattern to convert.

        Returns
        -------
        RDKitMol
            RDKit molecule.
        """
        return Chem.MolFromSmarts(pattern)


class SmilesFilter(_BasePatternsFilter):
    """Filter to keep or remove molecules based on SMILES patterns.

    In contrast to the SMARTSFilter, which also can match SMILES, the SmilesFilter
    sanitizes the molecules and, e.g. checks kekulized bonds for aromaticity and
    then sets it to aromatic while the SmartsFilter detects alternating single and
    double bonds.

    Notes
    -----
    There are four possible scenarios:
        - mode = "any" & keep_matches = True: Needs to match at least one filter element.
        - mode = "any" & keep_matches = False: Must not match any filter element.
        - mode = "all" & keep_matches = True: Needs to match all filter elements.
        - mode = "all" & keep_matches = False: Must not match all filter elements.
    """

    def _pattern_to_mol(self, pattern: str) -> RDKitMol:
        """Convert SMILES pattern to RDKit molecule.

        Parameters
        ----------
        pattern: str
            SMILES pattern to convert.

        Returns
        -------
        RDKitMol
            RDKit molecule.
        """
        return Chem.MolFromSmiles(pattern)


class ComplexFilter(_BaseKeepMatchesFilter):
    """Filter to keep or remove molecules based on multiple filter elements.

    Attributes
    ----------
    pipeline_filter_elements: Sequence[tuple[str, _MolToMolPipelineElement]]
        pairs of unique names and MolToMol elements to use as filters.
    [...]

    Notes
    -----
    There are four possible scenarios:
        - mode = "any" & keep_matches = True: Needs to match at least one filter element.
        - mode = "any" & keep_matches = False: Must not match any filter element.
        - mode = "all" & keep_matches = True: Needs to match all filter elements.
        - mode = "all" & keep_matches = False: Must not match all filter elements.
    """

    _filter_elements: Mapping[str, tuple[int, Optional[int]]]

    def __init__(
        self,
        pipeline_filter_elements: Sequence[tuple[str, _MolToMolPipelineElement]],
        keep_matches: bool = True,
        mode: FilterModeType = "any",
        name: str | None = None,
        n_jobs: int = 1,
        uuid: str | None = None,
    ) -> None:
        """Initialize ComplexFilter.

        Parameters
        ----------
        pipeline_filter_elements: Sequence[tuple[str, _MolToMolPipelineElement]]
            Filter elements to use.
        keep_matches: bool, optional (default: True)
            If True, keep matches, else remove matches.
        mode: FilterModeType, optional (default: "any")
            Mode to filter by.
        name: str, optional (default: None)
            Name of the pipeline element.
        n_jobs: int, optional (default: 1)
            Number of parallel jobs to use.
        uuid: str, optional (default: None)
            Unique identifier of the pipeline element.
        """
        self.pipeline_filter_elements = pipeline_filter_elements
        super().__init__(
            filter_elements=pipeline_filter_elements,
            keep_matches=keep_matches,
            mode=mode,
            name=name,
            n_jobs=n_jobs,
            uuid=uuid,
        )

    def get_params(self, deep: bool = True) -> dict[str, Any]:
        """Get parameters of ComplexFilter.

        Parameters
        ----------
        deep: bool, optional (default: True)
            If True, return the parameters of all subobjects that are PipelineElements.

        Returns
        -------
        dict[str, Any]
            Parameters of ComplexFilter.
        """
        params = super().get_params(deep)
        params.pop("filter_elements")
        params["pipeline_filter_elements"] = self.pipeline_filter_elements
        if deep:
            for name, element in self.pipeline_filter_elements:
                deep_items = element.get_params().items()
                params.update(
                    ("pipeline_filter_elements" + "__" + name + "__" + key, val)
                    for key, val in deep_items
                )
        return params

    def set_params(self, **parameters: Any) -> Self:
        """Set parameters of ComplexFilter.

        Parameters
        ----------
        parameters: Any
            Parameters to set.

        Returns
        -------
        Self
            Self.
        """
        parameter_copy = dict(parameters)
        if "pipeline_filter_elements" in parameter_copy:
            self.pipeline_filter_elements = parameter_copy.pop(
                "pipeline_filter_elements"
            )
            self.filter_elements = self.pipeline_filter_elements  # type: ignore
        for key in parameters:
            if key.startswith("pipeline_filter_elements__"):
                value = parameter_copy.pop(key)
                element_name, element_key = key.split("__")[1:]
                for name, element in self.pipeline_filter_elements:
                    if name == element_name:
                        element.set_params(**{element_key: value})
        super().set_params(**parameter_copy)
        return self

    @property
    def filter_elements(
        self,
    ) -> Mapping[str, tuple[int, Optional[int]]]:
        """Get filter elements."""
        return self._filter_elements

    @filter_elements.setter
    def filter_elements(
        self,
        filter_elements: Sequence[tuple[str, _MolToMolPipelineElement]],
    ) -> None:
        """Set filter elements.

        Parameters
        ----------
        filter_elements: Sequence[tuple[str, _MolToMolPipelineElement]]
            Filter elements to set.
        """
        self.filter_elements_dict = dict(filter_elements)
        if not len(self.filter_elements_dict) == len(filter_elements):
            raise ValueError("Filter elements names need to be unique.")
        self._filter_elements = {
            element_name: (1, None) for element_name, _element in filter_elements
        }

    def _calculate_single_element_value(
        self, filter_element: Any, value: RDKitMol
    ) -> int:
        """Calculate a single filter match for a molecule.

        Parameters
        ----------
        filter_element: Any
            MolToMol Filter to calculate.
        value: RDKitMol
            Molecule to calculate filter match for.

        Returns
        -------
        int
            Filter match.
        """
        mol = self.filter_elements_dict[filter_element].pretransform_single(value)
        if isinstance(mol, InvalidInstance):
            return 0
        return 1


class RDKitDescriptorsFilter(_BaseKeepMatchesFilter):
    """Filter to keep or remove molecules based on RDKit descriptors.

    Attributes
    ----------
    filter_elements: dict[str, FloatCountRange]
        Dictionary of RDKit descriptors to filter by.
        The value must be a tuple of minimum and maximum. If None, no limit is set.
    [...]

    Notes
    -----
    There are four possible scenarios:
        - mode = "any" & keep_matches = True: Needs to match at least one filter element.
        - mode = "any" & keep_matches = False: Must not match any filter element.
        - mode = "all" & keep_matches = True: Needs to match all filter elements.
        - mode = "all" & keep_matches = False: Must not match all filter elements.
    """

    @property
    def filter_elements(self) -> dict[str, FloatCountRange]:
        """Get allowed descriptors as dict."""
        return self._filter_elements

    @filter_elements.setter
    def filter_elements(self, descriptors: dict[str, FloatCountRange]) -> None:
        """Set allowed descriptors as dict.

        Parameters
        ----------
        descriptors: dict[str, FloatCountRange]
            Dictionary of RDKit descriptors to filter by.
        """
        if not all(hasattr(Descriptors, descriptor) for descriptor in descriptors):
            raise ValueError(
                "You are trying to use an invalid descriptor. Use RDKit Descriptors module."
            )
        self._filter_elements = descriptors

    def _calculate_single_element_value(
        self, filter_element: Any, value: RDKitMol
    ) -> float:
        """Calculate a single descriptor value for a molecule.

        Parameters
        ----------
        filter_element: Any
            Descriptor to calculate.
        value: RDKitMol
            Molecule to calculate descriptor for.

        Returns
        -------
        float
            Descriptor value.
        """
        return getattr(Descriptors, filter_element)(value)


class MixtureFilter(_MolToMolPipelineElement):
    """MolToMol which removes molecules composed of multiple fragments."""

    def pretransform_single(self, value: RDKitMol) -> OptionalMol:
        """Invalidate molecule containing multiple fragments.

        Parameters
        ----------
        value: RDKitMol
            Molecule to check.

        Returns
        -------
        OptionalMol
            Molecule if it contains only one fragment, else InvalidInstance.
        """
        fragments = Chem.GetMolFrags(value, asMols=True)
        if len(fragments) > 1:
            smiles_fragments = [Chem.MolToSmiles(fragment) for fragment in fragments]
            return InvalidInstance(
                self.uuid,
                f"Molecule contains multiple fragments: {' '.join(smiles_fragments)}",
                self.name,
            )
        return value


class EmptyMoleculeFilter(_MolToMolPipelineElement):
    """EmptyMoleculeFilter which removes empty molecules."""

    def pretransform_single(self, value: RDKitMol) -> OptionalMol:
        """Invalidate empty molecule.

        Parameters
        ----------
        value: RDKitMol
            Molecule to check.

        Returns
        -------
        OptionalMol
            Molecule if it is not empty, else InvalidInstance.
        """
        if value.GetNumAtoms() == 0:
            return InvalidInstance(self.uuid, "Molecule contains no atoms.", self.name)
        return value


class InorganicsFilter(_MolToMolPipelineElement):
    """Filters Molecules which do not contain any organic (i.e. Carbon) atoms."""

    CARBON_INORGANICS = ["O=C=O", "[C-]#[O+]"]  # CO2 and CO are not organic
    CARBON_INORGANICS_MAX_ATOMS = 3

    def pretransform_single(self, value: RDKitMol) -> OptionalMol:
        """Invalidate molecules not containing a carbon atom.

        Parameters
        ----------
        value: RDKitMol
            Molecule to check.

        Returns
        -------
        OptionalMol
            Molecule if it contains carbon, else InvalidInstance.
        """
        if not any(atom.GetAtomicNum() == 6 for atom in value.GetAtoms()):
            return InvalidInstance(
                self.uuid, "Molecule contains no organic atoms.", self.name
            )

        # Only check for inorganic molecules if the molecule is small enough
        if value.GetNumAtoms() <= self.CARBON_INORGANICS_MAX_ATOMS:
            smiles = Chem.MolToSmiles(value)
            if smiles in self.CARBON_INORGANICS:
                return InvalidInstance(self.uuid, "Molecule is not organic.", self.name)
        return value



"""Murcko scaffold clustering estimator."""

from __future__ import annotations

from numbers import Integral
from typing import Any, Literal

import numpy as np
import numpy.typing as npt
from sklearn.base import BaseEstimator, ClusterMixin, _fit_context
from sklearn.preprocessing import FunctionTransformer, OrdinalEncoder
from sklearn.utils.validation import validate_data

from molpipeline import ErrorFilter, FilterReinserter, Pipeline
from molpipeline.any2mol import AutoToMol
from molpipeline.mol2any import MolToSmiles
from molpipeline.mol2mol import EmptyMoleculeFilter, MakeScaffoldGeneric, MurckoScaffold
from molpipeline.utils.molpipeline_types import AnyStep, OptionalMol

try:
    from typing import Self  # type: ignore[attr-defined]
except ImportError:
    from typing_extensions import Self


__all__ = [
    "MurckoScaffoldClustering",
]


class MurckoScaffoldClustering(ClusterMixin, BaseEstimator):
    """Murcko scaffold clustering estimator."""

    linear_molecules_strategy: Literal["ignore", "own_cluster"]
    labels_: npt.NDArray[np.int32]
    n_clusters_: int
    _parameter_constraints: dict[str, Any] = {
        "n_jobs": [Integral, None],
    }
    use_smiles: bool

    def __init__(
        self,
        *,
        make_generic: bool = False,
        n_jobs: int = 1,
        linear_molecules_strategy: Literal["ignore", "own_cluster"] = "ignore",
    ):
        """Initialize Murcko scaffold clustering estimator.

        Parameters
        ----------
        make_generic : bool (default=False)
            Makes a Murcko scaffold generic (i.e. all atom types->C and all bonds->single).
        n_jobs : int, optional (default=1)
            Number of jobs to use for parallelization.
        linear_molecules_strategy : Literal["ignore", "own_cluster"], optional (default="ignore")
            Strategy for handling linear molecules. Can be "ignore" or "own_cluster". "ignore" will
            ignore linear molecules, and they will be replaced with NaN in the resulting clustering.
            "own_cluster" will instead cluster linear molecules in their own cluster and give them
            a valid cluster label.
        """
        self.n_jobs = n_jobs
        self.linear_molecules_strategy = linear_molecules_strategy
        self.make_generic = make_generic

    def _generate_pipeline(self) -> Pipeline:
        """Generate the pipeline for the Murcko scaffold clustering estimator.

        Returns
        -------
        Pipeline
            Pipeline for the Murcko scaffold clustering estimator.
        """
        auto2mol = AutoToMol()
        empty_mol_filter1 = EmptyMoleculeFilter()
        murcko_scaffold = MurckoScaffold()
        empty_mol_filter2 = EmptyMoleculeFilter()
        mol2smi = MolToSmiles()

        pipeline_step_list: list[AnyStep] = [
            ("auto2mol", auto2mol),
            ("empty_mol_filter1", empty_mol_filter1),
            ("murcko_scaffold", murcko_scaffold),
        ]

        scaffold_generic_elem: MakeScaffoldGeneric | None = None
        if self.make_generic:
            scaffold_generic_elem = MakeScaffoldGeneric()
            pipeline_step_list.append(("make_scaffold_generic", scaffold_generic_elem))

        pipeline_step_list.extend(
            [
                ("empty_mol_filter2", empty_mol_filter2),
                ("mol2smi", mol2smi),
            ]
        )

        if self.linear_molecules_strategy == "ignore":
            error_filter = ErrorFilter(filter_everything=True)
            pipeline_step_list.append(("error_filter", error_filter))

        elif self.linear_molecules_strategy == "own_cluster":
            # Create error filter for all errors except empty_mol_filter2
            # This is needed to give linear molecules (empty scaffold) a valid cluster label
            filter_ele_list = [auto2mol, empty_mol_filter1, murcko_scaffold]
            if scaffold_generic_elem is not None:
                filter_ele_list.append(scaffold_generic_elem)
            filter_ele_list.append(mol2smi)
            error_filter = ErrorFilter.from_element_list(filter_ele_list)
            pipeline_step_list.append(("error_filter", error_filter))

            # Create and add separate error replacer for murcko_scaffold
            no_scaffold_filter = ErrorFilter.from_element_list([empty_mol_filter2])
            no_scaffold_replacer = FilterReinserter.from_error_filter(
                no_scaffold_filter, "linear"
            )

            # Directly add the error filter and replacer to the pipeline
            pipeline_step_list.append(("no_scaffold_filter", no_scaffold_filter))
            pipeline_step_list.append(("no_scaffold_replacer", no_scaffold_replacer))
        else:
            raise ValueError(
                f"Invalid value for linear_molecules_strategy: {self.linear_molecules_strategy}"
            )

        error_replacer = FilterReinserter.from_error_filter(error_filter, np.nan)

        pipeline_step_list.extend(
            [
                ("reshape2d", FunctionTransformer(func=np.atleast_2d)),
                ("transpose", FunctionTransformer(func=np.transpose)),
                ("scaffold_encoder", OrdinalEncoder()),
                ("reshape1d", FunctionTransformer(func=np.ravel)),
                ("error_replacer", error_replacer),
            ]
        )

        cluster_pipeline = Pipeline(
            pipeline_step_list,
            n_jobs=self.n_jobs,
        )
        return cluster_pipeline

    # pylint: disable=C0103,W0613
    @_fit_context(prefer_skip_nested_validation=True)
    def fit(
        self,
        X: npt.NDArray[np.str_] | list[str] | list[OptionalMol],
        y: npt.NDArray[np.float64] | None = None,
        **params: Any,
    ) -> Self:
        """Fit Murcko scaffold clustering estimator.

        Parameters
        ----------
        X : array-like of shape (n_samples,)
            Smiles or molecule list or array.
        y : Ignored
            Not used, present for API consistency by convention.
        **params : Any
            Additional keyword arguments.

        Returns
        -------
        Self
            Fitted estimator.
        """
        X = validate_data(self, X=X, ensure_min_samples=2, ensure_2d=False, dtype=None)
        return self._fit(X)

    # pylint: disable=C0103,W0613
    def _fit(
        self,
        X: npt.NDArray[np.str_] | list[str] | list[OptionalMol],
    ) -> Self:
        """Fit Murcko scaffold clustering estimator.

        Parameters
        ----------
        X : array-like of shape (n_samples,)
            Smiles or molecule list or array.

        Returns
        -------
        Self
            Fitted estimator.
        """
        cluster_pipeline = self._generate_pipeline()
        self.labels_ = cluster_pipeline.fit_transform(X, None)
        if self.labels_ is None:
            raise AssertionError("self.labels_ can not be None")
        self.n_clusters_ = len(np.unique(self.labels_[~np.isnan(self.labels_)]))
        return self

    def fit_predict(
        self,
        X: (
            npt.NDArray[np.str_] | list[str] | list[OptionalMol]
        ),  # pylint: disable=C0103
        y: npt.NDArray[np.float64] | None = None,
        **params: Any,  # pylint: disable=unused-argument
    ) -> npt.NDArray[np.float64]:
        """Fit and predict Murcko scaffold clustering estimator.

        Parameters
        ----------
        X : array-like of shape (n_samples,)
            Smiles or molecule list or array.
        y : Ignored
            Not used, present for API consistency by convention.
        **params : Any
            Additional keyword arguments.

        Returns
        -------
        np.ndarray[int]
            Cluster labels.
        """
        # pylint: disable=W0246
        return super().fit_predict(X, y, **params)



"""Module for testing if the lightning wrapper functions work as intended."""

import unittest

import lightning as pl

from molpipeline.estimators.chemprop.lightning_wrapper import (
    get_non_default_params_trainer,
    get_params_trainer,
)


class TestLightningWrapper(unittest.TestCase):
    """Test the lightning wrapper functions.

    Note
    ----
    These tests are not exhaustive.
    """

    def test_setting_deterministic(self) -> None:
        """Test setting the deterministic parameter."""
        trainer_params = get_params_trainer(pl.Trainer(deterministic=True))
        self.assertTrue(trainer_params["deterministic"])

        trainer_params = get_params_trainer(pl.Trainer(deterministic=False))
        self.assertFalse(trainer_params["deterministic"])

        trainer_params = get_non_default_params_trainer(pl.Trainer(deterministic=True))
        self.assertIn("deterministic", trainer_params)
        self.assertTrue(trainer_params["deterministic"])

        trainer_params = get_non_default_params_trainer(pl.Trainer(deterministic=False))
        # deterministic is by default False and hence will not be listed in the parameters
        self.assertNotIn("deterministic", trainer_params)


if __name__ == "__main__":
    unittest.main()



"""Defines the pipeline which handles pipeline elements for molecular operations."""

from __future__ import annotations

from typing import Any, Iterable, Union

try:
    from typing import Self  # type: ignore[attr-defined]
except ImportError:
    from typing_extensions import Self

import numpy as np
from joblib import Parallel, delayed
from rdkit.Chem.rdchem import MolSanitizeException
from rdkit.rdBase import BlockLogs

from molpipeline.abstract_pipeline_elements.core import (
    ABCPipelineElement,
    InvalidInstance,
    RemovedInstance,
    TransformingPipelineElement,
)
from molpipeline.error_handling import (
    ErrorFilter,
    FilterReinserter,
    _MultipleErrorFilter,
)
from molpipeline.utils.molpipeline_types import TypeFixedVarSeq
from molpipeline.utils.multi_proc import check_available_cores


class _MolPipeline:
    """Contains the PipeElements which describe the functionality of the pipeline."""

    _n_jobs: int
    _element_list: list[ABCPipelineElement]
    _requires_fitting: bool

    def __init__(
        self,
        element_list: list[ABCPipelineElement],
        n_jobs: int = 1,
        name: str = "MolPipeline",
    ) -> None:
        """Initialize MolPipeline.

        Parameters
        ----------
        element_list: list[ABCPipelineElement]
            List of Pipeline Elements which form the pipeline.
        n_jobs:
            Number of cores used.
        name: str
            Name of pipeline.

        Returns
        -------
        None
        """
        self._element_list = element_list
        self.n_jobs = n_jobs
        self.name = name
        self._requires_fitting = any(
            element.requires_fitting for element in self._element_list
        )

    @property
    def _filter_elements(self) -> list[ErrorFilter]:
        """Get the elements which filter the input."""
        return [
            element
            for element in self._element_list
            if isinstance(element, ErrorFilter)
        ]

    @property
    def _filter_elements_agg(self) -> _MultipleErrorFilter:
        """Get the aggregated filter element."""
        return _MultipleErrorFilter(self._filter_elements)

    @property
    def _transforming_elements(
        self,
    ) -> list[Union[TransformingPipelineElement, _MolPipeline]]:
        """Get the elements which transform the input."""
        return [
            element
            for element in self._element_list
            if isinstance(element, (TransformingPipelineElement, _MolPipeline))
        ]

    @property
    def n_jobs(self) -> int:
        """Return the number of cores to use in transformation step."""
        return self._n_jobs

    @n_jobs.setter
    def n_jobs(self, requested_jobs: int) -> None:
        """Set the number of cores to use in transformation step.

        Parameters
        ----------
        requested_jobs: int
            Number of cores requested for transformation steps.
            If fewer cores than requested are available, the number of cores is set to maximum available.

        Returns
        -------
        None
        """
        self._n_jobs = check_available_cores(requested_jobs)

    @property
    def parameters(self) -> dict[str, Any]:
        """Get all parameters defining the object."""
        return self.get_params()

    @parameters.setter
    def parameters(self, parameter_dict: dict[str, Any]) -> None:
        """Set parameters of the pipeline and pipeline elements.

        Parameters
        ----------
        parameter_dict: dict[str, Any]
            Dictionary containing the parameter names and corresponding values to be set.

        Returns
        -------
        None
        """
        self.set_params(**parameter_dict)

    @property
    def requires_fitting(self) -> bool:
        """Return whether the pipeline requires fitting."""
        return self._requires_fitting

    def get_params(self, deep: bool = True) -> dict[str, Any]:
        """Get all parameters defining the object.

        Parameters
        ----------
        deep: bool
            If True get a deep copy of the parameters.

        Returns
        -------
        dict[str, Any]
            Dictionary containing the parameter names and corresponding values.
        """
        if deep:
            return {
                "element_list": self.element_list,
                "n_jobs": self.n_jobs,
                "name": self.name,
            }
        return {
            "element_list": self._element_list,
            "n_jobs": self.n_jobs,
            "name": self.name,
        }

    def set_params(self, **parameter_dict: Any) -> Self:
        """Set parameters of the pipeline and pipeline elements.

        Parameters
        ----------
        parameter_dict: Any
            Dictionary containing the parameter names and corresponding values to be set.

        Returns
        -------
        Self
            MolPipeline object with updated parameters.
        """
        if "element_list" in parameter_dict:
            self._element_list = parameter_dict["element_list"]
        if "n_jobs" in parameter_dict:
            self.n_jobs = int(parameter_dict["n_jobs"])
        if "name" in parameter_dict:
            self.name = str(parameter_dict["name"])
        return self

    @property
    def element_list(self) -> list[ABCPipelineElement]:
        """Get a shallow copy from the list of pipeline elements."""
        return self._element_list[:]  # [:] to create shallow copy.

    def _get_meta_element_list(
        self,
    ) -> list[Union[ABCPipelineElement, _MolPipeline]]:
        """Merge elements which do not require fitting to a meta element which improves parallelization.

        Returns
        -------
        list[Union[ABCPipelineElement, _MolPipeline]]
            List of pipeline elements and meta elements.
        """
        meta_element_list: list[Union[ABCPipelineElement, _MolPipeline]] = []
        no_fit_element_list: list[ABCPipelineElement] = []
        for element in self._element_list:
            if (
                isinstance(element, TransformingPipelineElement)
                and not element.requires_fitting
            ):
                no_fit_element_list.append(element)
            else:
                if len(no_fit_element_list) == 1:
                    meta_element_list.append(no_fit_element_list[0])
                elif len(no_fit_element_list) > 1:
                    meta_element_list.append(
                        _MolPipeline(no_fit_element_list, n_jobs=self.n_jobs)
                    )
                no_fit_element_list = []
                meta_element_list.append(element)
        if len(no_fit_element_list) == 1:
            meta_element_list.append(no_fit_element_list[0])
        elif len(no_fit_element_list) > 1:
            meta_element_list.append(
                _MolPipeline(no_fit_element_list, n_jobs=self.n_jobs)
            )
        return meta_element_list

    def fit(
        self,
        x_input: Any,
        y: Any = None,  # pylint: disable=invalid-name
        **fit_params: dict[Any, Any],
    ) -> Self:
        """Fit the MolPipeline according to x_input.

        Parameters
        ----------
        x_input: Any
            Molecular representations which are subsequently processed.
        y: Any
            Optional label of input. Only for SKlearn compatibility.
        fit_params: Any
            Parameters. Only for SKlearn compatibility.

        Returns
        -------
        Self
            Fitted MolPipeline.
        """
        _ = y  # Making pylint happy
        _ = fit_params  # Making pylint happy
        if self.requires_fitting:
            self.fit_transform(x_input)
        return self

    def fit_transform(  # pylint: disable=invalid-name,unused-argument
        self,
        x_input: Any,
        y: Any = None,
        **fit_params: dict[str, Any],
    ) -> Any:
        """Fit the MolPipeline according to x_input and return the transformed molecules.

        Parameters
        ----------
        x_input: Any
            Molecular representations which are subsequently processed.
        y: Any
            Optional label of input. Only for SKlearn compatibility.
        fit_params: Any
            Parameters. Only for SKlearn compatibility.

        Returns
        -------
        Any
            Transformed molecules.
        """
        iter_input = x_input

        removed_rows: dict[ErrorFilter, list[int]] = {}
        for error_filter in self._filter_elements:
            removed_rows[error_filter] = []
        iter_idx_array = np.arange(len(iter_input))

        # The meta elements merge steps which do not require fitting to improve parallelization
        for i_element in self._get_meta_element_list():
            if not isinstance(i_element, (TransformingPipelineElement, _MolPipeline)):
                continue
            i_element.n_jobs = self.n_jobs
            iter_input = i_element.pretransform(iter_input)
            for error_filter in self._filter_elements:
                iter_input = error_filter.transform(iter_input)
                for idx in error_filter.error_indices:
                    idx = iter_idx_array[idx]
                    removed_rows[error_filter].append(idx)
                iter_idx_array = error_filter.co_transform(iter_idx_array)
            if i_element.requires_fitting:
                if isinstance(i_element, _MolPipeline):
                    raise AssertionError("No subpipline should require fitting!")
                i_element.fit_to_result(iter_input)
            if isinstance(i_element, TransformingPipelineElement):
                iter_input = i_element.finalize_list(iter_input)
            iter_input = i_element.assemble_output(iter_input)
            i_element.n_jobs = 1

        # Set removed rows to filter elements to allow for correct co_transform
        iter_idx_array = np.arange(len(x_input))
        for error_filter in self._filter_elements:
            removed_idx_list = removed_rows[error_filter]
            error_filter.error_indices = []
            for new_idx, _idx in enumerate(iter_idx_array):
                if _idx in removed_idx_list:
                    error_filter.error_indices.append(new_idx)
            error_filter.n_total = len(iter_idx_array)
            iter_idx_array = error_filter.co_transform(iter_idx_array)
        error_replacer_list = [
            ele for ele in self._element_list if isinstance(ele, FilterReinserter)
        ]
        for error_replacer in error_replacer_list:
            error_replacer.select_error_filter(self._filter_elements)
            iter_input = error_replacer.transform(iter_input)
        return iter_input

    def transform_single(self, input_value: Any) -> Any:
        """Transform a single input according to the sequence of provided PipelineElements.

        Parameters
        ----------
        input_value: Any
            Molecular representation which is subsequently transformed.

        Returns
        -------
        Any
            Transformed molecular representation.
        """
        log_block = BlockLogs()
        iter_value = input_value
        for p_element in self._element_list:
            try:
                if not isinstance(iter_value, RemovedInstance):
                    iter_value = p_element.transform_single(iter_value)
                elif isinstance(p_element, FilterReinserter):
                    iter_value = p_element.transform_single(iter_value)
            except MolSanitizeException as err:
                iter_value = InvalidInstance(
                    p_element.uuid,
                    f"RDKit MolSanitizeException: {err.args}",
                    p_element.name,
                )
        del log_block
        return iter_value

    def pretransform(self, x_input: Any) -> Any:
        """Transform the input according to the sequence BUT skip the assemble output step.

        Parameters
        ----------
        x_input: Any
            Molecular representations which are subsequently transformed.

        Returns
        -------
        Any
            Transformed molecular representations.
        """
        return list(self._transform_iterator(x_input))

    def transform(self, x_input: Any) -> Any:
        """Transform the input according to the sequence of provided PipelineElements.

        Parameters
        ----------
        x_input: Any
            Molecular representations which are subsequently transformed.

        Returns
        -------
        Any
            Transformed molecular representations.
        """
        output_generator = self._transform_iterator(x_input)
        return self.assemble_output(output_generator)

    def assemble_output(self, value_list: Iterable[Any]) -> Any:
        """Assemble the output of the pipeline.

        Parameters
        ----------
        value_list: Iterable[Any]
            Generator which yields the output of the pipeline.

        Returns
        -------
        Any
            Assembled output.
        """
        last_element = self._transforming_elements[-1]
        if hasattr(last_element, "assemble_output"):
            return last_element.assemble_output(value_list)
        return list(value_list)

    def _finish(self) -> None:
        """Inform each pipeline element that the iterations have finished."""
        for p_element in self._element_list:
            p_element.finish()

    def _transform_iterator(self, x_input: Any) -> Any:
        """Transform the input according to the sequence of provided PipelineElements.

        Parameters
        ----------
        x_input: Any
            Molecular representations which are subsequently transformed.

        Returns
        -------
        Any
            Transformed molecular representations.
        """
        agg_filter = self._filter_elements_agg
        for filter_element in self._filter_elements:
            filter_element.error_indices = []
        parallel = Parallel(
            n_jobs=self.n_jobs,
            return_as="generator",
            batch_size="auto",
        )
        output_generator = parallel(
            delayed(self.transform_single)(value) for value in x_input
        )
        for i, transformed_value in enumerate(output_generator):
            if isinstance(transformed_value, RemovedInstance):
                agg_filter.register_removed(i, transformed_value)
            else:
                yield transformed_value
        agg_filter.set_total(len(x_input))
        self._finish()

    def co_transform(self, x_input: TypeFixedVarSeq) -> TypeFixedVarSeq:
        """Filter flagged rows from the input.

        Parameters
        ----------
        x_input: Any
            Molecular representations which are subsequently filtered.

        Returns
        -------
        Any
            Filtered molecular representations.
        """
        return self._filter_elements_agg.co_transform(x_input)



"""Test MolFilter, which invalidate molecules based on criteria defined in the respective filter."""

import json
import tempfile
import unittest
from pathlib import Path

from molpipeline import ErrorFilter, FilterReinserter, Pipeline
from molpipeline.any2mol import SmilesToMol
from molpipeline.mol2any import MolToSmiles
from molpipeline.mol2mol import (
    ComplexFilter,
    ElementFilter,
    InorganicsFilter,
    MixtureFilter,
    RDKitDescriptorsFilter,
    SmartsFilter,
    SmilesFilter,
)
from molpipeline.utils.comparison import compare_recursive
from molpipeline.utils.json_operations import recursive_from_json, recursive_to_json
from molpipeline.utils.molpipeline_types import FloatCountRange, IntOrIntCountRange

# pylint: disable=duplicate-code  # test case molecules are allowed to be duplicated
SMILES_ANTIMONY = "[SbH6+3]"
SMILES_BENZENE = "c1ccccc1"
SMILES_CHLOROBENZENE = "Clc1ccccc1"
SMILES_CL_BR = "NC(Cl)(Br)C(=O)O"
SMILES_METAL_AU = "OC[C@H]1OC(S[Au])[C@H](O)[C@@H](O)[C@@H]1O"

SMILES_LIST = [
    SMILES_ANTIMONY,
    SMILES_BENZENE,
    SMILES_CHLOROBENZENE,
    SMILES_METAL_AU,
    SMILES_CL_BR,
]


class ElementFilterTest(unittest.TestCase):
    """Unittest for Elementiflter."""

    def test_element_filter(self) -> None:
        """Test if molecules are filtered correctly by allowed chemical elements."""
        default_atoms_dict = {
            1: (0, None),
            5: (0, None),
            6: (0, None),
            7: (0, None),
            8: (0, None),
            9: (0, None),
            14: (0, None),
            15: (0, None),
            16: (0, None),
            17: (0, None),
            34: (0, None),
            35: (0, None),
            53: (0, None),
        }

        element_filter = ElementFilter()
        self.assertEqual(element_filter.allowed_element_numbers, default_atoms_dict)
        pipeline = Pipeline(
            [
                ("Smiles2Mol", SmilesToMol()),
                ("ElementFilter", element_filter),
                ("Mol2Smiles", MolToSmiles()),
                ("ErrorFilter", ErrorFilter()),
            ],
        )

        test_params_list_with_results = [
            {
                "params": {},
                "result": [SMILES_BENZENE, SMILES_CHLOROBENZENE, SMILES_CL_BR],
            },
            {
                "params": {
                    "ElementFilter__allowed_element_numbers": {
                        6: 6,
                        1: (5, 6),
                        17: (0, 1),
                    }
                },
                "result": [SMILES_BENZENE, SMILES_CHLOROBENZENE],
            },
            {"params": {"ElementFilter__add_hydrogens": False}, "result": []},
        ]

        for test_params in test_params_list_with_results:
            pipeline.set_params(**test_params["params"])
            filtered_smiles = pipeline.fit_transform(SMILES_LIST)
            self.assertEqual(filtered_smiles, test_params["result"])

    def test_json_roundtrip(self) -> None:
        """Test if ElementFilter can be serialized and deserialized.

        Notes
        -----
        It is important to save the ElementFilter as a JSON file and then load it back.
        This is because json.dumps() sets the keys of the dictionary to strings.
        """
        element_filter = ElementFilter()
        json_object = recursive_to_json(element_filter)
        with tempfile.TemporaryDirectory() as temp_folder:
            temp_file_path = Path(temp_folder) / "test.json"
            with open(temp_file_path, "w", encoding="UTF-8") as out_file:
                json.dump(json_object, out_file)
            with open(temp_file_path, "r", encoding="UTF-8") as in_file:
                loaded_json_object = json.load(in_file)
        recreated_element_filter = recursive_from_json(loaded_json_object)

        original_params = element_filter.get_params()
        recreated_params = recreated_element_filter.get_params()
        self.assertEqual(original_params.keys(), recreated_params.keys())
        for param_name, original_value in original_params.items():
            with self.subTest(param_name=param_name):
                self.assertTrue(
                    compare_recursive(original_value, recreated_params[param_name]),
                    f"Original: {original_value}, Recreated: {recreated_params[param_name]}",
                )


class ComplexFilterTest(unittest.TestCase):
    """Unittest for ComplexFilter."""

    @staticmethod
    def _create_pipeline() -> Pipeline:
        """Create a pipeline with a complex filter.

        Returns
        -------
        Pipeline
            Pipeline with a complex filter.
        """
        element_filter_1 = ElementFilter({6: 6, 1: 6})
        element_filter_2 = ElementFilter({6: 6, 1: 5, 17: 1})

        multi_element_filter = ComplexFilter(
            (
                ("element_filter_1", element_filter_1),
                ("element_filter_2", element_filter_2),
            )
        )

        pipeline = Pipeline(
            [
                ("Smiles2Mol", SmilesToMol()),
                ("MultiElementFilter", multi_element_filter),
                ("Mol2Smiles", MolToSmiles()),
                ("ErrorFilter", ErrorFilter()),
            ],
        )
        return pipeline

    def test_complex_filter(self) -> None:
        """Test if molecules are filtered correctly by allowed chemical elements."""
        pipeline = ComplexFilterTest._create_pipeline()

        test_params_list_with_results = [
            {
                "params": {},
                "result": [SMILES_BENZENE, SMILES_CHLOROBENZENE],
            },
            {
                "params": {"MultiElementFilter__mode": "all"},
                "result": [],
            },
            {
                "params": {
                    "MultiElementFilter__mode": "any",
                    "MultiElementFilter__pipeline_filter_elements__element_filter_1__add_hydrogens": False,
                },
                "result": [SMILES_CHLOROBENZENE],
            },
        ]

        for test_params in test_params_list_with_results:
            pipeline.set_params(**test_params["params"])
            filtered_smiles = pipeline.fit_transform(SMILES_LIST)
            self.assertEqual(filtered_smiles, test_params["result"])

    def test_json_serialization(self) -> None:
        """Test if complex filter can be serialized and deserialized."""
        pipeline = ComplexFilterTest._create_pipeline()
        json_object = recursive_to_json(pipeline)
        newpipeline = recursive_from_json(json_object)
        self.assertEqual(json_object, recursive_to_json(newpipeline))

        pipeline_result = pipeline.fit_transform(SMILES_LIST)
        newpipeline_result = newpipeline.fit_transform(SMILES_LIST)
        self.assertEqual(pipeline_result, newpipeline_result)

    def test_complex_filter_non_unique_names(self) -> None:
        """Test if molecules are filtered correctly by allowed chemical elements."""
        element_filter_1 = ElementFilter({6: 6, 1: 6})
        element_filter_2 = ElementFilter({6: 6, 1: 5, 17: 1})

        with self.assertRaises(ValueError):
            ComplexFilter(
                (("filter_1", element_filter_1), ("filter_1", element_filter_2))
            )


class SmartsSmilesFilterTest(unittest.TestCase):
    """Unittest for SmartsFilter and SmilesFilter."""

    def test_smarts_smiles_filter(self) -> None:
        """Test if molecules are filtered correctly by allowed SMARTS patterns."""
        smarts_pats: dict[str, IntOrIntCountRange] = {
            "c": (4, None),
            "Cl": 1,
        }
        smarts_filter = SmartsFilter(smarts_pats)

        smiles_pats: dict[str, IntOrIntCountRange] = {
            "c1ccccc1": (1, None),
            "Cl": 1,
        }
        smiles_filter = SmilesFilter(smiles_pats)

        for filter_ in [smarts_filter, smiles_filter]:
            new_input_as_list = list(filter_.filter_elements.keys())
            pipeline = Pipeline(
                [
                    ("Smiles2Mol", SmilesToMol()),
                    ("SmartsFilter", filter_),
                    ("Mol2Smiles", MolToSmiles()),
                    ("ErrorFilter", ErrorFilter()),
                ],
            )

            test_params_list_with_results = [
                {
                    "params": {},
                    "result": [SMILES_BENZENE, SMILES_CHLOROBENZENE, SMILES_CL_BR],
                },
                {
                    "params": {"SmartsFilter__keep_matches": False},
                    "result": [SMILES_ANTIMONY, SMILES_METAL_AU],
                },
                {
                    "params": {
                        "SmartsFilter__mode": "all",
                        "SmartsFilter__keep_matches": True,
                    },
                    "result": [SMILES_CHLOROBENZENE],
                },
                {
                    "params": {
                        "SmartsFilter__keep_matches": True,
                        "SmartsFilter__filter_elements": ["I"],
                    },
                    "result": [],
                },
                {
                    "params": {
                        "SmartsFilter__keep_matches": False,
                        "SmartsFilter__mode": "any",
                        "SmartsFilter__filter_elements": new_input_as_list,
                    },
                    "result": [SMILES_ANTIMONY, SMILES_METAL_AU],
                },
            ]

            for test_params in test_params_list_with_results:
                pipeline.set_params(**test_params["params"])
                filtered_smiles = pipeline.fit_transform(SMILES_LIST)
                self.assertEqual(filtered_smiles, test_params["result"])

    def test_smarts_smiles_filter_wrong_pattern(self) -> None:
        """Test if molecules are filtered correctly by allowed SMARTS patterns."""
        smarts_pats: dict[str, IntOrIntCountRange] = {
            "cIOnk": (4, None),
            "cC": 1,
        }
        with self.assertRaises(ValueError):
            SmartsFilter(smarts_pats)

        smiles_pats: dict[str, IntOrIntCountRange] = {
            "cC": (1, None),
            "Cl": 1,
        }
        with self.assertRaises(ValueError):
            SmilesFilter(smiles_pats)

    def test_smarts_filter_parallel(self) -> None:
        """Test if molecules are filtered correctly by allowed SMARTS patterns in parallel."""
        smarts_pats: dict[str, IntOrIntCountRange] = {
            "c": (4, None),
            "Cl": 1,
            "cc": (1, None),
            "ccc": (1, None),
            "cccc": (1, None),
            "ccccc": (1, None),
            "cccccc": (1, None),
            "c1ccccc1": (1, None),
            "cCl": 1,
        }
        smarts_filter = SmartsFilter(smarts_pats, mode="all", n_jobs=2)
        pipeline = Pipeline(
            [
                ("Smiles2Mol", SmilesToMol()),
                ("SmartsFilter", smarts_filter),
                ("Mol2Smiles", MolToSmiles()),
                ("ErrorFilter", ErrorFilter()),
            ],
        )
        filtered_smiles = pipeline.fit_transform(SMILES_LIST)
        self.assertEqual(filtered_smiles, [SMILES_CHLOROBENZENE])


class RDKitDescriptorsFilterTest(unittest.TestCase):
    """Unittest for RDKitDescriptorsFilter."""

    def test_descriptor_filter(self) -> None:
        """Test if molecules are filtered correctly by allowed descriptors."""
        descriptors: dict[str, FloatCountRange] = {
            "MolWt": (None, 190),
            "NumHAcceptors": (2, 10),
        }

        descriptor_filter = RDKitDescriptorsFilter(descriptors)

        pipeline = Pipeline(
            [
                ("Smiles2Mol", SmilesToMol()),
                ("DescriptorsFilter", descriptor_filter),
                ("Mol2Smiles", MolToSmiles()),
                ("ErrorFilter", ErrorFilter()),
            ],
        )
        test_params_list_with_results = [
            {"params": {}, "result": SMILES_LIST},
            {"params": {"DescriptorsFilter__mode": "all"}, "result": [SMILES_CL_BR]},
            {
                "params": {"DescriptorsFilter__keep_matches": False},
                "result": [
                    SMILES_ANTIMONY,
                    SMILES_BENZENE,
                    SMILES_CHLOROBENZENE,
                    SMILES_METAL_AU,
                ],
            },
            {"params": {"DescriptorsFilter__mode": "any"}, "result": []},
            {
                "params": {
                    "DescriptorsFilter__keep_matches": True,
                    "DescriptorsFilter__filter_elements": {"NumHAcceptors": (2.00, 4)},
                },
                "result": [SMILES_CL_BR],
            },
            {
                "params": {
                    "DescriptorsFilter__filter_elements": {"NumHAcceptors": (1.99, 4)}
                },
                "result": [SMILES_CL_BR],
            },
            {
                "params": {
                    "DescriptorsFilter__filter_elements": {"NumHAcceptors": (2.01, 4)}
                },
                "result": [],
            },
            {
                "params": {
                    "DescriptorsFilter__filter_elements": {"NumHAcceptors": (1, 2.00)}
                },
                "result": [SMILES_CL_BR],
            },
            {
                "params": {
                    "DescriptorsFilter__filter_elements": {"NumHAcceptors": (1, 2.01)}
                },
                "result": [SMILES_CL_BR],
            },
            {
                "params": {
                    "DescriptorsFilter__filter_elements": {"NumHAcceptors": (1, 1.99)}
                },
                "result": [],
            },
        ]

        for test_params in test_params_list_with_results:
            pipeline.set_params(**test_params["params"])
            filtered_smiles = pipeline.fit_transform(SMILES_LIST)
            self.assertEqual(filtered_smiles, test_params["result"])


class MixtureFilterTest(unittest.TestCase):
    """Unittest for MixtureFilter."""

    def test_invalidate_mixtures(self) -> None:
        """Test if mixtures are correctly invalidated."""
        mol_list = ["CCC.CC.C", "c1ccccc1.[Na+].[Cl-]", "c1ccccc1"]
        expected_invalidated_mol_list = [None, None, "c1ccccc1"]

        smi2mol = SmilesToMol()
        mixture_filter = MixtureFilter()
        mol2smi = MolToSmiles()
        error_filter = ErrorFilter.from_element_list([smi2mol, mixture_filter, mol2smi])
        error_replacer = FilterReinserter.from_error_filter(error_filter, None)

        pipeline = Pipeline(
            [
                ("smi2mol", smi2mol),
                ("mixture_filter", mixture_filter),
                ("mol2smi", mol2smi),
                ("error_filter", error_filter),
                ("error_replacer", error_replacer),
            ]
        )
        mols_processed = pipeline.fit_transform(mol_list)
        self.assertEqual(expected_invalidated_mol_list, mols_processed)


class InorganicsFilterTest(unittest.TestCase):
    """Unittest for InorganicsFilter."""

    def test_inorganic_filter(self) -> None:
        """Test if molecules are filtered correctly by allowed chemical elements."""
        smiles2mol = SmilesToMol()
        inorganics_filter = InorganicsFilter()
        mol2smiles = MolToSmiles()
        error_filter = ErrorFilter.from_element_list(
            [smiles2mol, inorganics_filter, mol2smiles]
        )
        pipeline = Pipeline(
            [
                ("Smiles2Mol", smiles2mol),
                ("ElementFilter", inorganics_filter),
                ("Mol2Smiles", mol2smiles),
                ("ErrorFilter", error_filter),
            ],
        )
        filtered_smiles = pipeline.fit_transform(SMILES_LIST)
        self.assertEqual(
            filtered_smiles,
            [SMILES_BENZENE, SMILES_CHLOROBENZENE, SMILES_METAL_AU, SMILES_CL_BR],
        )

        filtered_inroganics = pipeline.fit_transform(["O=C=O", "[O+]#[C-]"])
        self.assertEqual(
            filtered_inroganics,
            [],
        )

        filtered_inroganics = pipeline.fit_transform(InorganicsFilter.CARBON_INORGANICS)
        self.assertEqual(
            filtered_inroganics,
            [],
        )


if __name__ == "__main__":
    unittest.main()



"""Module for checking values."""

from __future__ import annotations

from typing import Any

__all__ = ["is_empty", "get_length"]


def is_empty(value: Any) -> bool:
    """Check if value is empty.

    Parameters
    ----------
    value: Any
        Value to be checked.

    Returns
    -------
    bool
        True if value is empty, False otherwise.

    Raises
    ------
    TypeError
        If value does not have a shape or len attribute.
    """
    if get_length(value) == 0:
        return True
    return False


def get_length(values: Any) -> int:
    """Get the length of the values as given by the shape or len attribute.

    Parameters
    ----------
    values: Any
        Values to be checked.

    Returns
    -------
    int
        Length of the values.
    """
    if hasattr(values, "shape"):
        return values.shape[0]
    if hasattr(values, "__len__"):
        return len(values)
    raise TypeError("Values must have a shape or len attribute.")



"""Classes for standardizing molecules."""

from __future__ import annotations

from typing import Any, Optional, Union

try:
    from typing import Self  # type: ignore[attr-defined]
except ImportError:
    from typing_extensions import Self

from rdkit import Chem
from rdkit.Chem import SaltRemover as rdkit_SaltRemover
from rdkit.Chem import SanitizeMol  # pylint: disable=no-name-in-module
from rdkit.Chem import rdMolHash, rdmolops
from rdkit.Chem.MolStandardize import rdMolStandardize

from molpipeline.abstract_pipeline_elements.core import (
    InvalidInstance,
)
from molpipeline.abstract_pipeline_elements.core import (
    MolToMolPipelineElement as _MolToMolPipelineElement,
)
from molpipeline.utils.molpipeline_types import OptionalMol, RDKitMol

MolHashing = Union[
    "rdMolHash.HashFunction.AnonymousGraph",  # type: ignore[valid-type]
    "rdMolHash.HashFunction.ArthorSubstructureOrder",  # type: ignore[valid-type]
    "rdMolHash.HashFunction.AtomBondCounts",  # type: ignore[valid-type]
    "rdMolHash.HashFunction.CanonicalSmiles",  # type: ignore[valid-type]
    "rdMolHash.HashFunction.DegreeVector",  # type: ignore[valid-type]
    "rdMolHash.HashFunction.ElementGraph",  # type: ignore[valid-type]
    "rdMolHash.HashFunction.ExtendedMurcko",  # type: ignore[valid-type]
    "rdMolHash.HashFunction.HetAtomProtomer",  # type: ignore[valid-type]
    "rdMolHash.HashFunction.HetAtomTautomer",  # type: ignore[valid-type]
    "rdMolHash.HashFunction.HetAtomTautomerv2",  # type: ignore[valid-type]
    "rdMolHash.HashFunction.Mesomer",  # type: ignore[valid-type]
    "rdMolHash.HashFunction.MolFormula",  # type: ignore[valid-type]
    "rdMolHash.HashFunction.MurckoScaffold",  # type: ignore[valid-type]
    "rdMolHash.HashFunction.NetCharge",  # type: ignore[valid-type]
    "rdMolHash.HashFunction.RedoxPair",  # type: ignore[valid-type]
    "rdMolHash.HashFunction.Regioisomer",  # type: ignore[valid-type]
    "rdMolHash.HashFunction.SmallWorldIndexBR",  # type: ignore[valid-type]
    "rdMolHash.HashFunction.SmallWorldIndexBRL",  # type: ignore[valid-type]
]


class TautomerCanonicalizer(_MolToMolPipelineElement):
    """MolToMolPipelineElement which canonicalizes tautomers of a molecule."""

    def pretransform_single(self, value: RDKitMol) -> OptionalMol:
        """Canonicalize tautomers of molecule.

        Parameters
        ----------
        value: RDKitMol
            Molecule to canonicalize tautomers from.

        Returns
        -------
        OptionalMol
            Canonicalized molecule if possible, else InvalidInstance.
        """
        enumerator = rdMolStandardize.TautomerEnumerator()
        try:
            return enumerator.Canonicalize(value)
        except RuntimeError as error:
            return InvalidInstance(
                self.uuid, f"Tautomer enumeration failed.: {error}", self.name
            )


class ChargeParentExtractor(_MolToMolPipelineElement):
    """MolToMolPipelineElement which returns charge-parent of a molecule, if possible.

    The charge-parent is the largest fragment after neutralization.
    """

    def pretransform_single(self, value: RDKitMol) -> OptionalMol:
        """Return charge-parent of molecule, which is the largest fragment after neutralization.

        Parameters
        ----------
        value: RDKitMol
            Molecule to remove charges from.

        Returns
        -------
        OptionalMol
            Charge-parent of molecule if possible, else InvalidInstance.
        """
        return rdMolStandardize.ChargeParent(value)


class FragmentDeduplicator(_MolToMolPipelineElement):
    """MolToMolPipelineElement which removes duplicate fragments from a molecule.

    Duplicates are detected by comparing the MolHashes of the fragments.
    """

    hashing_method: MolHashing

    def __init__(
        self,
        name: str = "FragmentDeduplicator",
        hashing_method: MolHashing = rdMolHash.HashFunction.HetAtomTautomer,
        n_jobs: int = 1,
        uuid: Optional[str] = None,
    ) -> None:
        """Initialize FragmentDeduplicator.

        Parameters
        ----------
        name: str, optional (default: "FragmentDeduplicator")
            Name of the pipeline element.
        hashing_method: MolHashing, optional (default: rdMolHash.HashFunction.HetAtomTautomer)
            MolHashing method to use for comparing fragments.
        n_jobs: int, optional (default: 1)
            Number of parallel jobs to use.
        uuid: str, optional (default: None)
            Unique identifier of the pipeline element.

        Returns
        -------
        None
        """
        super().__init__(name=name, n_jobs=n_jobs, uuid=uuid)
        self.hashing_method = hashing_method

    def pretransform_single(self, value: RDKitMol) -> OptionalMol:
        """Remove duplicate fragments from molecule.

        Parameters
        ----------
        value: RDKitMol
            Molecule to remove duplicate fragments from.

        Returns
        -------
        OptionalMol
            Molecule without duplicate fragments if possible, else InvalidInstance.
        """
        fragments = Chem.GetMolFrags(value, asMols=True)
        fragment_hash_list = [
            (rdMolHash.MolHash(fragment, self.hashing_method), fragment)
            for fragment in fragments
        ]
        if len(fragment_hash_list) == 0:
            return InvalidInstance(
                self.uuid, "Molecule contains no fragments.", self.name
            )
        unique_fragment_hashes = {fragment_hash_list[0][0]}
        recombined_fragment = fragment_hash_list[0][1]
        for fragment_hash, fragment in fragment_hash_list[1:]:
            if fragment_hash not in unique_fragment_hashes:
                unique_fragment_hashes.add(fragment_hash)
                recombined_fragment = Chem.CombineMols(recombined_fragment, fragment)

        for dict_key, dict_value in value.GetPropsAsDict(includeComputed=False).items():
            recombined_fragment.SetProp(dict_key, dict_value)
        return recombined_fragment


class LargestFragmentChooser(_MolToMolPipelineElement):
    """MolToMolPipelineElement which returns the largest fragment of a molecule."""

    def pretransform_single(self, value: RDKitMol) -> OptionalMol:
        """Return largest fragment of molecule.

        Parameters
        ----------
        value: RDKitMol
            Molecule to remove charges from.

        Returns
        -------
        OptionalMol
            Largest fragment of molecule if possible, else InvalidInstance.
        """
        return rdMolStandardize.LargestFragmentChooser().choose(value)


class MetalDisconnector(_MolToMolPipelineElement):
    """MolToMolPipelineElement which removes bonds between organic compounds and metals."""

    def pretransform_single(self, value: RDKitMol) -> OptionalMol:
        """Cleave bonds with metals.

        Parameters
        ----------
        value: RDKitMol
            Molecule to disconnect metals from.

        Returns
        -------
        OptionalMol
            Molecule without bonds to metals if possible, else InvalidInstance.
        """
        mol = rdMolStandardize.MetalDisconnector().Disconnect(value)
        if mol is not None:
            # sometimes the molecule is not sanitized after disconnecting, e.g. RingInfo is not updated.
            SanitizeMol(mol)
        return mol


class IsotopeRemover(_MolToMolPipelineElement):
    """MolToMolPipelineElement which removes isotope information of atoms in a molecule."""

    def pretransform_single(self, value: RDKitMol) -> OptionalMol:
        """Remove isotope information of each atom.

        Attention: Explicit hydrogen atoms are not made implicit.
        You may need to use the ExplicitHydrogenRemover afterward.

        Parameters
        ----------
        value: RDKitMol
            Molecule to remove charges from.

        Returns
        -------
        OptionalMol
            Largest fragment of molecule if possible, else InvalidInstance.
        """
        for atom in value.GetAtoms():
            atom.SetIsotope(0)
        return value


class ExplicitHydrogenRemover(_MolToMolPipelineElement):
    """MolToMolPipelineElement which removes explicit hydrogen atoms from a molecule."""

    def pretransform_single(self, value: RDKitMol) -> OptionalMol:
        """Remove explicit hydrogen atoms.

        Parameters
        ----------
        value: RDKitMol
            Molecule to remove explicit hydrogen atoms from.

        Returns
        -------
        OptionalMol
            Molecule without explicit hydrogen atoms if possible, else InvalidInstance.
        """
        return Chem.RemoveHs(value)


class StereoRemover(_MolToMolPipelineElement):
    """MolToMolPipelineElement which removes stereo-information from the molecule."""

    def pretransform_single(self, value: RDKitMol) -> OptionalMol:
        """Remove stereo-information in molecule.

        Parameters
        ----------
        value: RDKitMol
            Molecule to remove stereo-information from.

        Returns
        -------
        OptionalMol
            Molecule without stereo-information if possible, else InvalidInstance.
        """
        copy_mol = RDKitMol(value)
        rdmolops.RemoveStereochemistry(copy_mol)
        return copy_mol


class SaltRemover(_MolToMolPipelineElement):
    """MolToMolPipelineElement which removes metal ions from molecule."""

    def pretransform_single(self, value: RDKitMol) -> OptionalMol:
        """Remove metal ions.

        Parameters
        ----------
        value: RDKitMol
            Molecule to remove metal ions from.

        Returns
        -------
        OptionalMol
            Molecule without metal ions if possible, else InvalidInstance.
        """
        salt_less_mol = rdkit_SaltRemover.SaltRemover().StripMol(value)
        return salt_less_mol


class SolventRemover(_MolToMolPipelineElement):
    """MolToMolPipelineElement which removes defined fragments from a molecule."""

    _solvent_mol_list: list[RDKitMol]
    standard_solvent_smiles_list: list[str] = [
        "[OH2]",
        "ClCCl",
        "ClC(Cl)Cl",
        "CCOC(=O)C",
        "CO",
        "CC(C)O",
        "CC(=O)C",
        "CS(=O)C",
        "CCO",
        "CN(C)C",
    ]

    def __init__(
        self,
        solvent_smiles_list: Optional[list[str]] = None,
        name: str = "SolventRemover",
        n_jobs: int = 1,
        uuid: Optional[str] = None,
    ) -> None:
        """Initialize SolventRemover.

        Taken from ChEMBL structure pipeline:
        https://github.com/chembl/ChEMBL_Structure_Pipeline/blob/master/chembl_structure_pipeline/data/solvents.smi

        Parameters
        ----------
        solvent_smiles_list: list[str], optional (default=None)
            List of SMILES of fragments to remove, by default None which uses the default solvent list:
            - WATER	[OH2]
            - DICHLOROMETHANE	ClCCl
            - TRICHLOROMETHANE	ClC(Cl)Cl
            - ETHYL ACETATE	CCOC(=O)C
            - METHANOL	CO
            - PROPAN-2-OL	CC(C)O
            - ACETONE	CC(=O)C
            - DMSO	CS(=O)C
            - ETHANOL	CCO
            - Trimethylamine	CN(C)C  # Not in ChEMBL list
        name: str, optional (default="SolventRemover")
            Name of the pipeline element.
        n_jobs: int, optional (default=1)
            Number of parallel jobs to use.
        uuid: str, optional (default=None)
            Unique identifier of the pipeline element. If None, a random UUID is generated.
        """
        super().__init__(name=name, n_jobs=n_jobs, uuid=uuid)
        if solvent_smiles_list is None:
            solvent_smiles_list = self.standard_solvent_smiles_list
        self.solvent_smiles_list = solvent_smiles_list

    @property
    def solvent_mol_list(self) -> list[RDKitMol]:
        """Return molecule representation of smiles list."""
        return self._solvent_mol_list

    @property
    def solvent_smiles_list(self) -> list[str]:
        """Return the smiles list."""
        return self._solvent_smiles_list

    @solvent_smiles_list.setter
    def solvent_smiles_list(self, solvent_smiles_list: list[str]) -> None:
        """Set the smiles list.

        Parameters
        ----------
        solvent_smiles_list: list[str]
            List of SMILES of fragments to remove.

        Returns
        -------
        None
        """
        self._solvent_smiles_list = solvent_smiles_list
        solvent_mol_list = []
        for smiles in solvent_smiles_list:
            mol = Chem.MolFromSmiles(smiles)
            if mol is None:
                raise ValueError(f"Could not convert {smiles} to a molecule.")
            solvent_mol_list.append(mol)
        self._solvent_mol_list = solvent_mol_list

    def get_params(self, deep: bool = True) -> dict[str, Any]:
        """Return parameters of pipeline element.

        Parameters
        ----------
        deep: bool, optional (default: True)
            If True, return a deep copy of the parameters, else only a shallow copy.

        Returns
        -------
        dict[str, Any]
            Parameters of pipeline element.
        """
        params = super().get_params(deep=deep)
        if deep:
            params["solvent_smiles_list"] = [
                str(smi) for smi in self.solvent_smiles_list
            ]
        else:
            params["solvent_smiles_list"] = self.solvent_smiles_list
        return params

    def set_params(self, **parameters: Any) -> Self:
        """Set parameters of pipeline element.

        Parameters
        ----------
        parameters: Any
            Parameters to set.

        Returns
        -------
        Self
            Pipeline element with set parameters.
        """
        param_copy = dict(parameters)
        solvent_smiles_list = param_copy.pop("solvent_smiles_list", None)
        if solvent_smiles_list is not None:
            self.solvent_smiles_list = solvent_smiles_list
        super().set_params(**param_copy)
        return self

    def pretransform_single(self, value: RDKitMol) -> OptionalMol:
        """Remove all fragments from molecule.

        Parameters
        ----------
        value: RDKitMol
            Molecule to remove fragments from.

        Returns
        -------
        OptionalMol
            Molecule without fragments if possible, else InvalidInstance.
        """
        kept_fragments = []
        for fragment in Chem.GetMolFrags(value, asMols=True):
            n_atoms_f = fragment.GetNumAtoms()
            n_bonds_f = fragment.GetNumBonds()
            for solvent_mol in self._solvent_mol_list:
                if (
                    n_atoms_f == solvent_mol.GetNumAtoms()
                    and n_bonds_f == solvent_mol.GetNumBonds()
                    and fragment.HasSubstructMatch(solvent_mol)
                ):
                    break  # if it matches, stop searching
            else:  # no break: no match
                kept_fragments.append(fragment)
        if len(kept_fragments) == 0:
            return InvalidInstance(self.uuid, "All fragments were removed.", self.name)
        combined_fragments = kept_fragments[0]
        for fragment in kept_fragments[1:]:
            combined_fragments = Chem.CombineMols(combined_fragments, fragment)
        for dict_key, dict_value in value.GetPropsAsDict(includeComputed=False).items():
            combined_fragments.SetProp(dict_key, dict_value)
        return combined_fragments


class Uncharger(_MolToMolPipelineElement):
    """MolToMolPipelineElement which removes charges in a molecule, if possible."""

    def pretransform_single(self, value: RDKitMol) -> OptionalMol:
        """Remove charges of molecule.

        Parameters
        ----------
        value: RDKitMol
            Molecule to remove charges from.

        Returns
        -------
        OptionalMol
            Uncharged molecule if possible, else InvalidInstance.
        """
        return rdMolStandardize.Uncharger().uncharge(value)
